'''
/*
 * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013 Nicira, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* OpenFlow: protocol between controller and datapath. */
Created on 2015/7/13

:author: hubo
'''
from .common import *
from . import common
from namedstruct.namedstruct import StructDefWarning
import warnings as _warnings

with _warnings.catch_warnings():
    _warnings.filterwarnings('ignore', '^padding', StructDefWarning)
    
    '''
    /* Port number(s)   meaning
     * ---------------  --------------------------------------
     * 0x0000           not assigned a meaning by OpenFlow 1.0
     * 0x0001...0xfeff  "physical" ports
     * 0xff00...0xfff7  "reserved" but not assigned a meaning by OpenFlow 1.0
     * 0xfff8...0xffff  "reserved" OFPP_* ports with assigned meanings
     */
    
    /* Ranges. */
    '''
    ofp_port = enum('ofp_port',
                       globals(),
                       uint16,
                        OFPP_MAX = 0xff00, # /* Max # of switch ports. */
                        
                        # /* Reserved output "ports". */
                        OFPP_IN_PORT = 0xfff8, # /* Where the packet came in. */
                        OFPP_TABLE = 0xfff9, # /* Perform actions in flow table. */
                        OFPP_NORMAL = 0xfffa, # /* Process with normal L2/L3. */
                        OFPP_FLOOD = 0xfffb, # /* All ports except input port and
                                            #                        * ports disabled by STP. */
                        OFPP_ALL = 0xfffc, # /* All ports except input port. */
                        OFPP_CONTROLLER = 0xfffd, # /* Send to controller. */
                        OFPP_LOCAL = 0xfffe, # /* Local openflow "port". */
                        OFPP_NONE = 0xffff # /* Not associated with any port. */
    )
    
    ofp_port_no = ofp_port
    
    OFPP_FIRST_RESV = 0xfff8, # /* First assigned reserved port. */
    OFPP_LAST_RESV = 0xffff, # /* Last assigned reserved port. */
    
    
    ofp_type = ofp_type.extend(globals(),
        OFPT_VENDOR             = 4,
        OFPT_FEATURES_REQUEST   = 5,  #/* Controller/switch message */
        OFPT_FEATURES_REPLY     = 6,  #/* Controller/switch message */
        OFPT_GET_CONFIG_REQUEST = 7,  #/* Controller/switch message */
        OFPT_GET_CONFIG_REPLY   = 8,  #/* Controller/switch message */
        OFPT_SET_CONFIG         = 9,  #/* Controller/switch message */
    
        OFPT_PACKET_IN          = 10, #/* Async message */
        OFPT_FLOW_REMOVED       = 11, #/* Async message */
        OFPT_PORT_STATUS        = 12, #/* Async message */
    
        OFPT_PACKET_OUT         = 13, #/* Controller/switch message */
        OFPT_FLOW_MOD           = 14, #/* Controller/switch message */
        OFPT_PORT_MOD           = 15, #/* Controller/switch message */
        OFPT_STATS_REQUEST      = 16, #/* Controller/switch message */
        OFPT_STATS_REPLY        = 17, #/* Controller/switch message */
    
        OFPT_BARRIER_REQUEST    = 18, #/* Controller/switch message */
        OFPT_BARRIER_REPLY      = 19, #/* Controller/switch message */
    
        OFPT_QUEUE_GET_CONFIG_REQUEST    = 20, #/* Controller/switch message */
        OFPT_QUEUE_GET_CONFIG_REPLY      = 21  #/* Controller/switch message */
    )
    
    ofp_type_reply_set = set([OFPT_ECHO_REPLY, OFPT_FEATURES_REPLY, OFPT_GET_CONFIG_REPLY, OFPT_STATS_REPLY, OFPT_BARRIER_REPLY, OFPT_QUEUE_GET_CONFIG_REPLY])
    
    ofp_type_asyncmessage_set = set([OFPT_PACKET_IN, OFPT_FLOW_REMOVED, OFPT_PORT_STATUS])
    
    OFP_VERSION = OFP10_VERSION
    
    ofp_msg = nstruct(name = 'ofp_msg',
                      base = common.ofp_msg_mutable,
                      criteria = lambda x: x.header.version == OFP_VERSION,
                      init = packvalue(OFP_VERSION, 'header', 'version'),
                      classifyby = (OFP_VERSION,),
                      classifier = lambda x: x.header.type,
                      extend = {('header','type') : ofp_type})
    
    ofp_vendor = nstruct((uint32, 'vendor'),
                         name = 'ofp_vendor',
                         base = ofp_msg,
                         criteria = lambda x: x.header.type == OFPT_VENDOR,
                         classifyby = (OFPT_VENDOR,),
                         init = packvalue(OFPT_VENDOR, 'header', 'type')
                         )
    
    ofp_error_type = ofp_error_type.extend(globals(),
                                           OFPET_FLOW_MOD_FAILED = 3,
                                           OFPET_PORT_MOD_FAILED = 4,
                                           OFPET_QUEUE_OP_FAILED = 5)
    
    '''
    /* ofp_error_msg 'code' values for OFPET_FLOW_MOD_FAILED. 'data' contains
    * at least the first 64 bytes of the failed request. */
    '''
    ofp_flow_mod_failed_code = enum('ofp_flow_mod_failed_code', globals(),
        OFPFMFC_ALL_TABLES_FULL = 0,        # /* Flow not added because of full tables. */
        OFPFMFC_OVERLAP = 1,                # /* Attempted to add overlapping flow with
    #                                          * CHECK_OVERLAP flag set. */
        OFPFMFC_EPERM = 2,                  # /* Permissions error. */
        OFPFMFC_BAD_EMERG_TIMEOUT = 3,      # /* Flow not added because of non-zero idle/hard
    #                                          * timeout. */
        OFPFMFC_BAD_COMMAND = 4,            # /* Unknown command. */
        OFPFMFC_UNSUPPORTED = 5,            # /* Unsupported action list - cannot process in
    #                                          * the order specified. */
    )
    
    '''
    /* ofp_error_msg 'code' values for OFPET_PORT_MOD_FAILED. 'data' contains
    * at least the first 64 bytes of the failed request. */
    '''
    ofp_port_mod_failed_code = enum('ofp_port_mod_failed_code', globals(),
        OFPPMFC_BAD_PORT = 0,       #/* Specified port does not exist. */
        OFPPMFC_BAD_HW_ADDR = 1,    #/* Specified hardware address is wrong. */
    )
    
    '''
    /* ofp_error msg 'code' values for OFPET_QUEUE_OP_FAILED. 'data' contains
    * at least the first 64 bytes of the failed request */
    '''
    ofp_queue_op_failed_code = enum('ofp_queue_op_failed_code', globals(),
        OFPQOFC_BAD_PORT = 0,   # /* Invalid port (or port does not exist). */
        OFPQOFC_BAD_QUEUE = 1,  # /* Queue does not exist. */
        OFPQOFC_EPERM = 2       # /* Permissions error. */
    )
    
    ofp_error_types = dict(ofp_error_types)
    
    ofp_error_types.update({OFPET_FLOW_MOD_FAILED : ofp_error_typedef(OFPET_FLOW_MOD_FAILED, ofp_flow_mod_failed_code, OFP_VERSION, ofp_error_type),
                            OFPET_PORT_MOD_FAILED : ofp_error_typedef(OFPET_PORT_MOD_FAILED, ofp_port_mod_failed_code, OFP_VERSION, ofp_error_type),
                            OFPET_QUEUE_OP_FAILED : ofp_error_typedef(OFPET_QUEUE_OP_FAILED, ofp_queue_op_failed_code, OFP_VERSION, ofp_error_type)})
    
    ofp_switch_config = nstruct((ofp_config_flags, 'flags'),
                                (uint16, 'miss_send_len'),
                                name = 'ofp_switch_config',
                                base = ofp_msg,
                                criteria = lambda x: x.header.type == OFPT_GET_CONFIG_REPLY or x.header.type == OFPT_SET_CONFIG,
                                classifyby = (OFPT_GET_CONFIG_REPLY, OFPT_SET_CONFIG),
                                init = packvalue(OFPT_SET_CONFIG, 'header','type'))
    
    
    
    '''
    /* OpenFlow 1.0 specific capabilities supported by the datapath (struct
     * ofp_switch_features, member capabilities). */
     '''
    ofp_capabilities  = ofp_capabilities.extend(
                               globals(),
                               OFPC_STP            = 1 << 3,  #/* 802.1d spanning tree. */
                               OFPC_RESERVED       = 1 << 4)  #/* Reserved, must not be set. */
    
    '''
    /* OpenFlow 1.0 specific current state of the physical port.  These are not
     * configurable from the controller.
     */
    /* The OFPPS10_STP_* bits have no effect on switch operation.  The
     * controller must adjust OFPPC_NO_RECV, OFPPC_NO_FWD, and
     * OFPPC_NO_PACKET_IN appropriately to fully implement an 802.1D spanning
     * tree. */
    '''
    ofp_port_state = ofp_port_state.extend(globals(),
        OFPPS_STP_LISTEN  = 0 << 8, # /* Not learning or relaying frames. */
        OFPPS_STP_LEARN   = 1 << 8, # /* Learning but not relaying frames. */
        OFPPS_STP_FORWARD = 2 << 8, # /* Learning and relaying frames. */
        OFPPS_STP_BLOCK   = 3 << 8  # /* Not part of spanning tree. */
        ) # /* Bit mask for OFPPS10_STP_* values. */
    
    OFPPS_STP_MASK    = 3 << 8
    
    OFPPS_ALL = OFPPS_LINK_DOWN | OFPPS_STP_MASK
    
    ofp_action_type = enum('ofp_action_type', globals(),
        uint16,
        OFPAT_OUTPUT = 0,             #/* Output to switch port. */
        OFPAT_SET_VLAN_VID = 1,       #/* Set the 802.1q VLAN id. */
        OFPAT_SET_VLAN_PCP = 2,       #/* Set the 802.1q priority. */
        OFPAT_STRIP_VLAN = 3,         #/* Strip the 802.1q header. */
        OFPAT_SET_DL_SRC = 4,         #/* Ethernet source address. */
        OFPAT_SET_DL_DST = 5,         #/* Ethernet destination address. */
        OFPAT_SET_NW_SRC = 6,         #/* IP source address. */
        OFPAT_SET_NW_DST = 7,         #/* IP destination address. */
        OFPAT_SET_NW_TOS = 8,         #/* IP ToS (DSCP field, 6 bits). */
        OFPAT_SET_TP_SRC = 9,         #/* TCP/UDP source port. */
        OFPAT_SET_TP_DST = 10,        #/* TCP/UDP destination port. */
        OFPAT_ENQUEUE = 11,           #/* Output to queue. */
        OFPAT_VENDOR = 0xffff)
    
    ofp_action = nstruct((ofp_action_type, 'type'),
                        (uint16, 'len'),
                        name = 'ofp_action',
                        size = sizefromlen(512, 'len'),
                        prepack = packsize('len'),
                        classifier = lambda x: x.type
                        )
    
    ofp_action_vendor = nstruct((uint32, 'vendor'),
                                       name = 'ofp_action_vendor',
                                       base = ofp_action,
                                       criteria = lambda x: x.type == OFPAT_VENDOR,
                                       classifyby = (OFPAT_VENDOR,),
                                       init = packvalue(OFPAT_VENDOR, 'type')
                                       )
    
    '''
    /* Action structure for OFPAT10_OUTPUT, which sends packets out 'port'.
     * When the 'port' is the OFPP_CONTROLLER, 'max_len' indicates the max
     * number of bytes to send.  A 'max_len' of zero means no bytes of the
     * packet should be sent. */
    '''
    ofp_action_output = nstruct((ofp_port, 'port'),
                                (uint16, 'max_len'),
                                name = 'ofp_action_output',
                                base = ofp_action,
                                criteria = lambda x: x.type == OFPAT_OUTPUT,
                                classifyby = (OFPAT_OUTPUT,),
                                init = packvalue(OFPAT_OUTPUT, 'type'))
    
    '''
    /* Action structure for OFPAT10_SET_VLAN_VID and OFPAT11_SET_VLAN_VID. */
    '''
    ofp_action_vlan_vid = nstruct(
        (uint16, 'vlan_vid'),             # /* VLAN id. */
        (uint8[2],),
        name = 'ofp_action_vlan_vid',
        base = ofp_action,
        criteria = lambda x: x.type == OFPAT_SET_VLAN_VID,
        classifyby = (OFPAT_SET_VLAN_VID,),
        init = packvalue(OFPAT_SET_VLAN_VID, 'type'))
    
    '''
    /* Action structure for OFPAT10_SET_VLAN_PCP and OFPAT11_SET_VLAN_PCP. */
    '''
    ofp_action_vlan_pcp = nstruct(
        (uint8, 'vlan_pcp'),            #  /* VLAN priority. */
        (uint8[3],),
        name = 'ofp_action_vlan_pcp',
        base = ofp_action,
        criteria = lambda x: x.type == OFPAT_SET_VLAN_PCP,
        classifyby = (OFPAT_SET_VLAN_PCP,),
        init = packvalue(OFPAT_SET_VLAN_PCP, 'type'))
    
    '''
    /* Action structure for OFPAT10_SET_DL_SRC/DST and OFPAT11_SET_DL_SRC/DST. */
    '''
    ofp_action_dl_addr = nstruct(
        (mac_addr, 'dl_addr'),  #  /* Ethernet address. */
        (uint8[6],),
        name = 'ofp_action_dl_addr',
        base = ofp_action,
        criteria = lambda x: x.type == OFPAT_SET_DL_SRC or x.type == OFPAT_SET_DL_DST,
        classifyby = (OFPAT_SET_DL_SRC, OFPAT_SET_DL_DST),
        init = packvalue(OFPAT_SET_DL_SRC, 'type'))
    
    '''
    /* Action structure for OFPAT10_SET_NW_SRC/DST and OFPAT11_SET_NW_SRC/DST. */
    '''
    ofp_action_nw_addr = nstruct(
        (ip4_addr, 'nw_addr'),               # /* IP address. */
        name = 'ofp_action_nw_addr',
        base = ofp_action,
        criteria = lambda x: x.type == OFPAT_SET_NW_SRC or x.type == OFPAT_SET_NW_DST,
        classifyby = (OFPAT_SET_NW_SRC, OFPAT_SET_NW_DST),
        init = packvalue(OFPAT_SET_NW_SRC, 'type'))
    
    '''
    /* Action structure for OFPAT10_SET_NW_TOS and OFPAT11_SET_NW_TOS. */
    '''
    ofp_action_nw_tos = nstruct(
        (uint8, 'nw_tos'),                 # /* DSCP in high 6 bits, rest ignored. */
        (uint8[3],),
        name = 'ofp_action_nw_tos',
        base = ofp_action,
        criteria = lambda x: x.type == OFPAT_SET_NW_TOS,
        classifyby = (OFPAT_SET_NW_TOS,),
        init = packvalue(OFPAT_SET_NW_TOS, 'type'))
    
    '''
    /* Action structure for OFPAT10_SET_TP_SRC/DST and OFPAT11_SET_TP_SRC/DST. */
    '''
    ofp_action_tp_port = nstruct(
        (uint16, 'tp_port'),               # /* TCP/UDP port. */
        (uint8[2],),
        name = 'ofp_action_tp_port',
        base = ofp_action,
        criteria = lambda x: x.type == OFPAT_SET_TP_SRC or x.type == OFPAT_SET_TP_DST,
        classifyby = (OFPAT_SET_TP_SRC, OFPAT_SET_TP_DST),
        init = packvalue(OFPAT_SET_TP_SRC, 'type'))
    
    
    '''
    /* OpenFlow 1.0 specific features of physical ports available in a datapath. */
    '''
    ofp_port_features = ofp_port_features.extend(globals(),
        OFPPF_COPPER     = 1 << 7,  #/* Copper medium. */
        OFPPF_FIBER      = 1 << 8,  #/* Fiber medium. */
        OFPPF_AUTONEG    = 1 << 9,  #/* Auto-negotiation. */
        OFPPF_PAUSE      = 1 << 10, #/* Pause. */
        OFPPF_PAUSE_ASYM = 1 << 11  #/* Asymmetric pause. */
    )
    
    '''
    /* Description of a physical port */
    '''
    ofp_phy_port = nstruct(
        (ofp_port, 'port_no'),
        (mac_addr, 'hw_addr'),
        (char[OFP_MAX_PORT_NAME_LEN], 'name'), #/* Null-terminated */
    
        (ofp_port_config, 'config'),     #   /* Bitmap of OFPPC_* and OFPPC10_* flags. */
        (ofp_port_state, 'state'),      #   /* Bitmap of OFPPS_* and OFPPS10_* flags. */
    
        #/* Bitmaps of OFPPF_* and OFPPF10_* that describe features.  All bits
        # * zeroed if unsupported or unavailable. */
        (ofp_port_features, 'curr'),       #   /* Current features. */
        (ofp_port_features, 'advertised'), #   /* Features being advertised by the port. */
        (ofp_port_features, 'supported'),  #   /* Features supported by the port. */
        (ofp_port_features, 'peer'),       #   /* Features advertised by peer. */
        name = 'ofp_phy_port'
    )
    
    ofp_action_type_bitwise = enum('ofp_action_type_bitwise', None, uint32, True,
                                   **dict((k, 1<<v) for (k,v) in ofp_action_type.getDict().items() if v < 32))
    
    ofp_switch_features = nstruct((uint64, 'datapath_id'),
                                  (uint32, 'n_buffers'),
                                  (uint8, 'n_tables'),
                                  (uint8[3],),
                                  (ofp_capabilities, 'capabilities'),
                                  (ofp_action_type_bitwise, 'actions'),
                                  (ofp_phy_port[0], 'ports'),
                                  name = 'ofp_switch_features',
                                  base = ofp_msg,
                                  criteria = lambda x: x.header.type == OFPT_FEATURES_REPLY,
                                  classifyby = (OFPT_FEATURES_REPLY,),
                                  init = packvalue(OFPT_FEATURES_REPLY, 'header', 'type'))
    
    
    '''
    /* Modify behavior of the physical port */
    '''
    ofp_port_mod = nstruct(
        (ofp_port, 'port_no'),
        (mac_addr, 'hw_addr'), 
    
        (ofp_port_config, 'config'),     #   /* Bitmap of OFPPC_* flags. */
        (ofp_port_config, 'mask'),       #   /* Bitmap of OFPPC_* flags to be changed. */
    
        (ofp_port_features, 'advertise'),  #   /* Bitmap of "ofp_port_features"s.  Zero all bits to prevent any action taking place. */
        (uint8[4],),            #   /* Pad to 64-bits. */
        name = 'ofp_port_mod',
        base = ofp_msg,
        criteria = lambda x: x.header.type == OFPT_PORT_MOD,
        classifyby = (OFPT_PORT_MOD,),
        init = packvalue(OFPT_PORT_MOD, 'header', 'type')
    )
    
    
    ofp_queue_prop_header = nstruct((ofp_queue_properties, 'property'),
                                    (uint16, 'len'),
                                    (uint8[4],),
                                    name = 'ofp_queue_prop_header')
    
    ofp_queue_prop = nstruct((ofp_queue_prop_header, 'prop_header'),
                             name = 'ofp_queue_prop',
                             size = sizefromlen(256, 'prop_header', 'len'),
                             prepack = packrealsize('prop_header', 'len'),
                             classifier = lambda x: x.prop_header.property
                             )
    
    ofp_queue_prop_min_rate = nstruct((uint16, 'rate'),
                                  (uint8[6],),
                                  base = ofp_queue_prop,
                                  criteria = lambda x: x.prop_header.property == OFPQT_MIN_RATE,
                                  classifyby = (OFPQT_MIN_RATE,),
                                  init = packvalue(OFPQT_MIN_RATE, 'prop_header', 'property'),
                                  name = 'ofp_queue_prop_min_rate')
    
    
    ofp_packet_queue = nstruct(
        (uint32, 'queue_id'),       #   /* id for the specific queue. */
        (uint16, 'len'),            #   /* Length in bytes of this queue desc. */
        (uint8[2],),                #   /* 64-bit alignment. */
        (ofp_queue_prop[0], 'properties'),
        name = 'ofp_packet_queue',
        size = sizefromlen(4096, 'len'),
        prepack = packsize('len')
        )
    '''
    /* Query for port queue configuration. */
    '''
    ofp_queue_get_config_request = nstruct(
        (uint16, 'port'),       #   /* Port to be queried. Should refer to a valid physical port (i.e. < OFPP_MAX) */
        (uint8[2],),
        name = 'ofp_queue_get_config_request',
        base = ofp_msg,
        criteria = lambda x: x.header.type == OFPT_QUEUE_GET_CONFIG_REQUEST,
        classifyby = (OFPT_QUEUE_GET_CONFIG_REQUEST,),
        init = packvalue(OFPT_QUEUE_GET_CONFIG_REQUEST, 'header', 'type')
    )
    
    '''
    /* Queue configuration for a given port. */
    '''
    ofp_queue_get_config_reply  = nstruct(
        (uint16, 'port'),
        (uint8[6],),
        (ofp_packet_queue[0], 'queues'),  # /* List of configured queues. */
        base = ofp_msg,
        name = 'ofp_queue_get_config_reply',
        criteria = lambda x: x.header.type == OFPT_QUEUE_GET_CONFIG_REPLY,
        classifyby = (OFPT_QUEUE_GET_CONFIG_REPLY,),
        init = packvalue(OFPT_QUEUE_GET_CONFIG_REPLY, 'header', 'type')
    )
    
    '''
    /* Packet received on port (datapath -> controller). */
    '''
    ofp_packet_in = nstruct(
        (uint32, 'buffer_id'),  #     /* ID assigned by datapath. */
        (uint16, 'total_len'),  #     /* Full length of frame. */
        (ofp_port, 'in_port'),    #     /* Port on which frame was received. */
        (ofp_packet_in_reason, 'reason'),      #     /* Reason packet is being sent (one of OFPR_*) */
        (uint8,),
        (raw, 'data'),          
        base = ofp_msg,
        name = 'ofp_packet_in',
        criteria = lambda x: x.header.type == OFPT_PACKET_IN,
        classifyby = (OFPT_PACKET_IN,),
        init = packvalue(OFPT_PACKET_IN, 'header', 'type')
    )
    
    '''
    /* OFPAT10_ENQUEUE action struct: send packets to given queue on port. */
    '''
    ofp_action_enqueue = nstruct(
        (uint16, 'port'),       #     /* Port that queue belongs. Should
        (uint8[6],),            #     /* Pad for 64-bit alignment. */
        (uint32, 'queue_id'),    #     /* Where to enqueue the packets. */
        name ='ofp_action_enqueue',
        base = ofp_action,
        criteria = lambda x: x.type == OFPAT_ENQUEUE,
        classifyby = (OFPAT_ENQUEUE,),
        init = packvalue(OFPAT_ENQUEUE, 'type')
    )
    
    '''
    /* Send packet (controller -> datapath). */
    '''
    
    def _ofp_packet_out_actions_packsize(x):
        x.actions_len = x._realsize() - 2
    ofp_packet_out_actions = nstruct(
        (uint16, 'actions_len'),
        (ofp_action[0], 'actions'),
        name = 'ofp_packet_out_actions',
        size = lambda x: x.actions_len + 2,
        prepack = _ofp_packet_out_actions_packsize,
        padding = 1)
    
    ofp_packet_out = nstruct(
        (uint32, 'buffer_id'),        #   /* ID assigned by datapath or UINT32_MAX. */
        (ofp_port, 'in_port'),          #   /* Packet's input port (OFPP_NONE if none). */
        (ofp_packet_out_actions,),
        (raw, 'data'),
        name = 'ofp_packet_out',
        base = ofp_msg,
        criteria = lambda x: x.header.type == OFPT_PACKET_OUT,
        classifyby = (OFPT_PACKET_OUT,),
        init = packvalue(OFPT_PACKET_OUT, 'header', 'type')
    )
    
    '''
    /* Flow wildcards. */
    '''
    
    OFPFW_NW_SRC_SHIFT = 8
    OFPFW_NW_SRC_BITS = 6
    
    OFPFW_NW_DST_SHIFT = 14
    OFPFW_NW_DST_BITS = 6
    
    
    ofp_flow_wildcards = enum('ofp_flow_wildcards', globals(),uint32,True,
        OFPFW_IN_PORT    = 1 << 0,  #/* Switch input port. */
        OFPFW_DL_VLAN    = 1 << 1,  #/* VLAN vid. */
        OFPFW_DL_SRC     = 1 << 2,  #/* Ethernet source address. */
        OFPFW_DL_DST     = 1 << 3,  #/* Ethernet destination address. */
        OFPFW_DL_TYPE    = 1 << 4,  #/* Ethernet frame type. */
        OFPFW_NW_PROTO   = 1 << 5,  #/* IP protocol. */
        OFPFW_TP_SRC     = 1 << 6,  #/* TCP/UDP source port. */
        OFPFW_TP_DST     = 1 << 7,  #/* TCP/UDP destination port. */
    #/* IP source address wildcard bit count.  0 is exact match, 1 ignores the
    #* LSB, 2 ignores the 2 least-significant bits, ..., 32 and higher wildcard
    #* the entire field.  This is the *opposite* of the usual convention where
    #* e.g. /24 indicates that 8 bits (not 24 bits) are wildcarded. */
        OFPFW_NW_SRC_MASK = (((1 << OFPFW_NW_SRC_BITS) - 1)
                               << OFPFW_NW_SRC_SHIFT),
        OFPFW_NW_SRC_ALL = 32 << OFPFW_NW_SRC_SHIFT,
    
    #    /* IP destination address wildcard bit count.  Same format as source. */
        OFPFW_NW_DST_MASK = (((1 << OFPFW_NW_DST_BITS) - 1)
                               << OFPFW_NW_DST_SHIFT),
        OFPFW_NW_DST_ALL = 32 << OFPFW_NW_DST_SHIFT,
    
        OFPFW_DL_VLAN_PCP = 1 << 20, # /* VLAN priority. */
        OFPFW_NW_TOS = 1 << 21, # /* IP ToS (DSCP field, 6 bits). */
    
    #    /* Wildcard all fields. */
        OFPFW_ALL = ((1 << 22) - 1)
    )
    
    
    
    #/* The wildcards for ICMP type and code fields use the transport source
    # * and destination port fields, respectively. */
    OFPFW_ICMP_TYPE = OFPFW_TP_SRC
    OFPFW_ICMP_CODE = OFPFW_TP_DST
    
    #/* The VLAN id is 12-bits, so we can use the entire 16 bits to indicate
    # * special conditions.  All ones indicates that 802.1Q header is not present.
    # */
    OFP_VLAN_NONE = 0xffff
    
    '''
    /* Fields to match against flows */
    '''
    ofp_match = nstruct(
        (ofp_flow_wildcards, 'wildcards'),    #    /* Wildcard fields. */
        (ofp_port, 'in_port'),      #    /* Input switch port. */
        (mac_addr, 'dl_src'),  #    /* Ethernet source address. */
        (mac_addr, 'dl_dst'), #    /* Ethernet destination address. */
        (uint16, 'dl_vlan'),      #    /* Input VLAN. */
        (uint8, 'dl_vlan_pcp'),   #    /* Input VLAN priority. */
        (uint8[1],),              #    /* Align to 64-bits. */
        (ethertype, 'dl_type'),      #    /* Ethernet frame type. */
        (uint8, 'nw_tos'),        #    /* IP ToS (DSCP field, 6 bits). */
        (uint8, 'nw_proto'),      #    /* IP protocol or lower 8 bits of ARP opcode. */
        (uint8[2],),              #    /* Align to 64-bits. */
        (ip4_addr, 'nw_src'),       #    /* IP source address. */
        (ip4_addr, 'nw_dst'),       #    /* IP destination address. */
        (uint16, 'tp_src'),       #    /* TCP/UDP source port. */
        (uint16, 'tp_dst'),       #    /* TCP/UDP destination port. */
        name = 'ofp_match'
    )
    
    ofp_flow_mod_flags = ofp_flow_mod_flags.extend(globals(),
        OFPFF_EMERG       = 1 << 2 #/* Part of "emergency flow cache". */
    )
    
    '''
    /* Flow setup and teardown (controller -> datapath). */
    '''
    ofp_flow_mod = nstruct(
        (ofp_match, 'match'),   #    /* Fields to match */
        (uint64, 'cookie'),     #    /* Opaque controller-issued identifier. */
    
    #    /* Flow actions. */
        (ofp_flow_mod_command, 'command'),    #         /* One of OFPFC_*. */
        (uint16, 'idle_timeout'),  #      /* Idle time before discarding (seconds). */
        (uint16, 'hard_timeout'),  #      /* Max time before discarding (seconds). */
        (uint16, 'priority'),      #      /* Priority level of flow entry. */
        (uint32, 'buffer_id'),     #      /* Buffered packet to apply to (or -1). Not meaningful for OFPFC_DELETE*. */
    #/* For OFPFC_DELETE* commands, require matching entries to include this as an
    # output port.  A value of OFPP_NONE indicates no restriction. */
        (ofp_port, 'out_port'),      
        (ofp_flow_mod_flags, 'flags'),         #      /* One of OFPFF_*. */
        (ofp_action[0], 'actions'),  #      /* The action length is inferred from the length field in the header. */
        base = ofp_msg,
        criteria = lambda x: x.header.type == OFPT_FLOW_MOD,
        classifyby = (OFPT_FLOW_MOD,),
        init = packvalue(OFPT_FLOW_MOD, 'header', 'type'),
        name = 'ofp_flow_mod'
    )
    
    '''
    /* Flow removed (datapath -> controller). */
    '''
    ofp_flow_removed = nstruct(
        (ofp_match, 'match'),   # /* Description of fields. */
        (uint64, 'cookie'),     # /* Opaque controller-issued identifier. */
    
        (uint16, 'priority'),   # /* Priority level of flow entry. */
        (ofp_flow_removed_reason, 'reason'),      # /* One of OFPRR_*. */
        (uint8[1],),            # /* Align to 32-bits. */
    
        (uint32, 'duration_sec'), #    /* Time flow was alive in seconds. */
        (uint32, 'duration_nsec'),#    /* Time flow was alive in nanoseconds beyond duration_sec. */
        (uint16, 'idle_timeout'), #    /* Idle timeout from original flow mod. */
        (uint8[2],),               #    /* Align to 64-bits. */
        (uint64, 'packet_count'),
        (uint64, 'byte_count'),
        name = 'ofp_flow_removed',
        base = ofp_msg,
        criteria = lambda x: x.header.type == OFPT_FLOW_REMOVED,
        classifyby = (OFPT_FLOW_REMOVED,),
        init = packvalue(OFPT_FLOW_REMOVED, 'header', 'type')
    )
    
    ofp_port_status = nstruct(
        (ofp_port_reason, 'reason'),
        (uint8[7],),
        (ofp_phy_port, 'desc'),
        name= 'ofp_port_status',
        base = ofp_msg,
        criteria = lambda x: x.header.type == OFPT_PORT_STATUS,
        classifyby = (OFPT_PORT_STATUS,),
        init = packvalue(OFPT_PORT_STATUS, 'header', 'type')
    )
    
    '''
    /* Statistics request or reply message. */
    '''
    
    ofp_stats_types = enum('ofp_stats_types', globals(),uint16,
    #/* Description of this OpenFlow switch.
    #* The request body is empty.
    #* The reply body is struct ofp_desc_stats. */
            OFPST_DESC = 0,
    #/* Individual flow statistics.
    #* The request body is struct ofp_flow_stats_request.
    #* The reply body is an array of struct ofp_flow_stats. */
            OFPST_FLOW = 1,
    #/* Aggregate flow statistics.
    #* The request body is struct ofp_aggregate_stats_request.
    #* The reply body is struct ofp_aggregate_stats_reply. */
            OFPST_AGGREGATE = 2,
    #/* Flow table statistics.
    #* The request body is empty.
    #* The reply body is an array of struct ofp_table_stats. */
            OFPST_TABLE = 3,
    #/* Physical port statistics.
    #* The request body is struct ofp_port_stats_request.
    #* The reply body is an array of struct ofp_port_stats. */
            OFPST_PORT = 4,
    #/* Queue statistics for a port
    #* The request body defines the port
    #* The reply body is an array of struct ofp_queue_stats */
            OFPST_QUEUE = 5,
    #/* Vendor extension.
    #* The request and reply bodies begin with a 32-bit vendor ID, which takes
    #* the same form as in "struct ofp_vendor_header". The request and reply
    #* bodies are otherwise vendor-defined. */
            OFPST_VENDOR = 0xffff
    )
    
    
    ofp_stats_msg = nstruct(
        (ofp_stats_types, 'type'),       #       /* One of the OFPST_* constants. */
        (ofp_stats_reply_flags, 'flags'),      #       /* Requests: always 0.
    #                                 * Replies: 0 or OFPSF_REPLY_MORE. */
        name = 'ofp_stats_msg',
        base = ofp_msg,
        criteria = lambda x: x.header.type == OFPT_STATS_REQUEST or x.header.type == OFPT_STATS_REPLY,
        classifyby = (OFPT_STATS_REQUEST, OFPT_STATS_REPLY),
        init = packvalue(OFPT_STATS_REQUEST, 'header', 'type')
    )
    
    ofp_stats_request = nstruct(
        name = 'ofp_stats_request',
        base = ofp_stats_msg,
        criteria = lambda x: x.header.type == OFPT_STATS_REQUEST,
        classifier = lambda x: x.type,
        init = packvalue(OFPT_STATS_REQUEST, 'header', 'type')
    )
    
    ofp_stats_reply = nstruct(
        name = 'ofp_stats_request',
        base = ofp_stats_msg,
        criteria = lambda x: x.header.type == OFPT_STATS_REPLY,
        classifier = lambda x: x.type,
        init = packvalue(OFPT_STATS_REPLY, 'header', 'type')
    )
    
    
    
    DESC_STR_LEN = 256
    SERIAL_NUM_LEN = 32
    
    ofp_desc_stats = nstruct((char[DESC_STR_LEN], 'mfr_desc'),
                       (char[DESC_STR_LEN], 'hw_desc'),
                       (char[DESC_STR_LEN], 'sw_desc'),
                       (char[SERIAL_NUM_LEN], 'serial_num'),
                       (char[DESC_STR_LEN], 'dp_desc'),
                       name = 'ofp_desc_stats')
    
    ofp_desc_stats_reply = nstruct(
                        (ofp_desc_stats,),
                        name = 'ofp_desc_stats_reply',
                        base = ofp_stats_reply,
                        criteria = lambda x: x.type == OFPST_DESC,
                        classifyby = (OFPST_DESC,),
                        init = packvalue(OFPST_DESC, 'type')
                                   )
    
    '''
    /* Stats request of type OFPST_AGGREGATE or OFPST_FLOW. */
    '''
    ofp_flow_stats_request = nstruct(
        (ofp_match, 'match'),   # /* Fields to match. */
        (ofp_table, 'table_id'),    # /* ID of table to read (from ofp_table_stats) or 0xff for all tables. */
        (uint8,),               # /* Align to 32 bits. */
        (ofp_port, 'out_port'),    # /* Require matching entries to include this as an output port.  A value of OFPP_NONE indicates no restriction. */
        name = 'ofp_flow_stats_request',
        base = ofp_stats_request,
        criteria = lambda x: x.type == OFPST_FLOW or x.type == OFPST_AGGREGATE,
        classifyby = (OFPST_FLOW, OFPST_AGGREGATE),
        init = packvalue(OFPST_FLOW, 'type')
    )
    
    '''
    /* Body of reply to OFPST_FLOW request. */
    '''
    ofp_flow_stats = nstruct(
        (uint16, 'length'),        #/* Length of this entry. */
        (uint8, 'table_id'),       #/* ID of table flow came from. */
        (uint8,),
        (ofp_match, 'match'),       #/* Description of fields. */
        (uint32, 'duration_sec'),  #/* Time flow has been alive in seconds. */
        (uint32, 'duration_nsec'), #/* Time flow has been alive in nanoseconds beyond duration_sec. */
        (uint16, 'priority'),      #/* Priority of the entry. Only meaningful when this is not an exact-match entry. */
        (uint16, 'idle_timeout'),  #/* Number of seconds idle before expiration. */
        (uint16, 'hard_timeout'),  #/* Number of seconds before expiration. */
        (uint8[6],),               #/* Align to 64 bits. */
        (uint64, 'cookie'),        #/* Opaque controller-issued identifier. */
        (uint64, 'packet_count'),  #/* Number of packets in flow. */
        (uint64, 'byte_count'),    #/* Number of bytes in flow. */
        (ofp_action[0], 'actions'),#/* Actions. */
        name = 'ofp_flow_stats',
        size = sizefromlen(4096, 'length'),
        prepack = packsize('length')
    )
    
    ofp_flow_stats_reply = nstruct(
        (ofp_flow_stats[0], 'stats'),
        name = 'ofp_flow_stats_reply',
        base = ofp_stats_reply,
        criteria = lambda x: x.type == OFPST_FLOW,
        classifyby = (OFPST_FLOW,),
        init = packvalue(OFPST_FLOW, 'type')
    )
    
    ofp_table = enum('ofp_table',
                     globals(),
                     uint8,
                     OFPTT_ALL = 0xff)
    
    '''
    /* Body for ofp_stats_request of type OFPST_AGGREGATE. */
    '''
    ofp_aggregate_stats_request = nstruct(
        (ofp_match, 'match'),             # /* Fields to match. */
        (ofp_table, 'table_id'),          # /* ID of table to read (from ofp_table_stats)
                                          #  0xff for all tables or 0xfe for emergency. */
        (uint8,),                         # /* Align to 32 bits. */
        (ofp_port, 'out_port'),           # /* Require matching entries to include this
                                          # as an output port. A value of OFPP_NONE
                                          # indicates no restriction. */
        base = ofp_stats_request,
        criteria = lambda x: x.type == OFPST_AGGREGATE,
        classifyby = (OFPST_AGGREGATE,),
        init = packvalue(OFPST_AGGREGATE, 'type'),
        name = 'ofp_aggregate_stats_request'
    )
    
    
    '''
    /* Body of reply to OFPST_AGGREGATE request. */
    '''
    ofp_aggregate_stats_reply = nstruct(
        (uint64, 'packet_count'),           # /* Number of packets in flows. */
        (uint64, 'byte_count'),             # /* Number of bytes in flows. */
        (uint32, 'flow_count'),             # /* Number of flows. */
        (uint8[4],),
        base = ofp_stats_reply,
        criteria = lambda x: x.type == OFPST_AGGREGATE,
        classifyby = (OFPST_AGGREGATE,),
        init = packvalue(OFPST_AGGREGATE, 'type'),
        name = 'ofp_aggregate_stats_reply'
    )
    
    
    '''
    /* Body of reply to OFPST_TABLE request. */
    '''
    ofp_table_stats = nstruct(
        (uint8, 'table_id'),        # /* Identifier of table.  Lower numbered tables are consulted first. */
        (uint8[3],),                # /* Align to 32-bits. */
        (char[OFP_MAX_TABLE_NAME_LEN], 'name'),
        (ofp_flow_wildcards, 'wildcards'),      # /* Bitmap of OFPFW_* wildcards that are supported by the table. */
        (uint32, 'max_entries'),    # /* Max number of entries supported. */
        (uint32, 'active_count'),   # /* Number of active entries. */
        (uint64, 'lookup_count'),   # /* # of packets looked up in table. */
        (uint64, 'matched_count'),   # /* Number of packets that hit table. */
        name = 'ofp_table_stats'
    )
    
    ofp_table_stats_reply = nstruct(
        (ofp_table_stats[0], 'stats'),
        name = 'ofp_table_stats_reply',
        base = ofp_stats_reply,
        criteria = lambda x: x.type == OFPST_TABLE,
        classifyby = (OFPST_TABLE,),
        init = packvalue(OFPST_TABLE, 'type')
    )
    
    '''
    /* Stats request of type OFPST_PORT. */
    '''
    ofp_port_stats_request = nstruct(
        (ofp_port, 'port_no'),
    #/* OFPST_PORT message may request statistics for a single port (specified with port_no)
    # or for all ports (port_no == OFPP_NONE). */
        (uint8[6],),
        name = 'ofp_port_stats_request',
        base = ofp_stats_request,
        criteria = lambda x: x.type == OFPST_PORT,
        classifyby = (OFPST_PORT,),
        init = packvalue(OFPST_PORT, 'type')
    )
    
    '''
    /* Body of reply to OFPST_PORT request. If a counter is unsupported, set
     * the field to all ones. */
     '''
    ofp_port_stats = nstruct(
        (uint16, 'port_no'),
        (uint8[6],),
        (uint64, 'rx_packets'),    # /* Number of received packets. */
        (uint64, 'tx_packets'),    # /* Number of transmitted packets. */
        (uint64, 'rx_bytes'),      # /* Number of received bytes. */
        (uint64, 'tx_bytes'),      # /* Number of transmitted bytes. */
        (uint64, 'rx_dropped'),    # /* Number of packets dropped by RX. */
        (uint64, 'tx_dropped'),    # /* Number of packets dropped by TX. */
        (uint64, 'rx_errors'),     # /* Number of receive errors.  This is a
    #super-set of receive errors and should be
    #great than or equal to the sum of all
    #rx_*_err values. */
        (uint64, 'tx_errors'),     # /* Number of transmit errors.  This is a super-set of transmit errors. */
        (uint64, 'rx_frame_err'),  # /* Number of frame alignment errors. */
        (uint64, 'rx_over_err'),   # /* Number of packets with RX overrun. */
        (uint64, 'rx_crc_err'),    # /* Number of CRC errors. */
        (uint64, 'collisions'),    # /* Number of collisions. */
        name = 'ofp_port_stats'
    )
    
    ofp_port_stats_reply = nstruct(
        (ofp_port_stats[0], 'stats'),
        name = 'ofp_port_stats_reply',
        base = ofp_stats_reply,
        criteria = lambda x: x.type == OFPST_PORT,
        classifyby = (OFPST_PORT,),
        init = packvalue(OFPST_PORT, 'type')
    )
    
    '''
    /* All ones is used to indicate all queues in a port (for stats retrieval). */
    '''
    ofp_queue = enum('ofp_queue', globals(), uint32,
    OFPQ_ALL = 0xffffffff)
    
    '''
    /* Body for stats request of type OFPST_QUEUE. */
    '''
    ofp_queue_stats_request = nstruct(
        (ofp_port, 'port_no'),     #   /* All ports if OFPP_ALL. */
        (uint8[2],),             #   /* Align to 32-bits. */
        (ofp_queue, 'queue_id'),    #   /* All queues if OFPQ_ALL. */
        name = 'ofp_queue_stats_request',
        base = ofp_stats_request,
        criteria = lambda x: x.type == OFPST_QUEUE,
        classifyby = (OFPST_QUEUE,),
        init = packvalue(OFPST_QUEUE, 'type')
    )
    
    '''
    /* Body for stats reply of type OFPST_QUEUE consists of an array of this
     * structure type. */
     '''
     
    ofp_queue_stats = nstruct(
        (uint16, 'port_no'),
        (uint8[2],),                # /* Align to 32-bits. */
        (uint32, 'queue_id'),       # /* Queue id. */
        (uint64, 'tx_bytes'),       # /* Number of transmitted bytes. */
        (uint64, 'tx_packets'),     # /* Number of transmitted packets. */
        (uint64, 'tx_errors'),      # /* # of packets dropped due to overrun. */
        name = 'ofp_queue_stats'
    )
    
    ofp_queue_stats_reply = nstruct(
        (ofp_queue_stats[0], 'stats'),
        name = 'ofp_queue_stats_reply',
        base = ofp_stats_reply,
        criteria = lambda x: x.type == OFPST_QUEUE,
        classifyby = (OFPST_QUEUE,),
        init = packvalue(OFPST_QUEUE, 'type')
    )
    
    '''
    /* Vendor extension stats message. */
    '''
    ofp_vendor_stats_request = nstruct(
        (uint32, 'vendor'),
        name = 'ofp_vendor_stats_request',
        base = ofp_stats_request,
        criteria = lambda x: x.type == OFPST_VENDOR,
        classifyby = (OFPST_VENDOR,),
        init = packvalue(OFPST_VENDOR, 'type')
    #    /* Followed by vendor-defined arbitrary additional data. */
    )
    
    ofp_vendor_stats_reply = nstruct(
        (uint32, 'vendor'),
        name = 'ofp_vendor_stats_reply',
        base = ofp_stats_reply,
        criteria = lambda x: x.type == OFPST_VENDOR,
        classifyby = (OFPST_VENDOR,),
        init = packvalue(OFPST_VENDOR, 'type')
    #    /* Followed by vendor-defined arbitrary additional data. */
    )
    
    ofp_vendor_vendorid = 'vendor'
    ofp_vendor_subtype = 'subtype'
    
    ofp_action_vendor_vendorid = 'vendor'
    ofp_action_vendor_subtype = 'subtype'
    
    ofp_stats_vendor_vendorid = 'vendor'
    ofp_stats_vendor_subtype = 'subtype'
    
    from .nicira_ext import *
    
    '''
    /* Header for Nicira vendor requests and replies. */
    '''
    nicira_header = nstruct(
        (nxt_subtype, 'subtype'),
        name = 'nicira_header',
        base = ofp_vendor,
        criteria = lambda x: x.vendor == NX_VENDOR_ID,
        init = packvalue(NX_VENDOR_ID, 'vendor'),
        classifier = lambda x: x.subtype
    )
    
    '''
    /* Header for Nicira-defined actions. */
    '''
    nx_action = nstruct(
        (nx_action_subtype, 'subtype'),     #               /* NXAST_*. */
        name = 'nx_action',
        base = ofp_action_vendor,
        criteria = lambda x: x.vendor == NX_VENDOR_ID,
        init = packvalue(NX_VENDOR_ID, 'vendor'),
        classifier = lambda x: x.subtype
    )
    
    nx_stats_request = nstruct(
        (nx_stats_subtype, 'subtype'),
        (uint8[4],),
        base = ofp_vendor_stats_request,
        criteria = lambda x: x.vendor == NX_VENDOR_ID,
        init = packvalue(NX_VENDOR_ID, 'vendor'),
        name = 'nx_stats_request',
        classifier = lambda x: getattr(x, 'subtype')
    )
    
    nx_stats_reply = nstruct(
        (nx_stats_subtype, 'subtype'),
        (uint8[4],),
        base = ofp_vendor_stats_reply,
        criteria = lambda x: x.vendor == NX_VENDOR_ID,
        init = packvalue(NX_VENDOR_ID, 'vendor'),
        name = 'nx_stats_reply',
        classifier = lambda x: getattr(x, 'subtype')
    )
    
    create_extension(globals(), nicira_header, nx_action, nx_stats_request, nx_stats_reply, ofp_vendor_subtype, ofp_action_vendor_subtype, ofp_stats_vendor_subtype)



from vlcp.config import defaultconfig
from vlcp.event import Event
from vlcp.event import RoutineContainer
from vlcp.event import withIndices
from vlcp.protocol.openflow import OpenflowConnectionStateEvent
from vlcp.server.module import callAPI, publicapi, api
from vlcp.service.sdn import ioprocessing
from vlcp.service.sdn import ovsdbportmanager
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.service.utils.remoteapi import remoteAPI
from vlcp.utils.ethernet import ETHERTYPE_8021Q


def _is_vxlan(obj):
    try:
        return obj.physicalnetwork.type == 'vxlan'
    except AttributeError:
        return False


@withIndices("connection","logicalnetworkid","type")
class VXLANMapChanged(Event):
    UPDATED = 'updated'
    DELETED = 'deleted'

@withIndices("connection","physname","phyiname","type")
class VtepControllerCall(Event):
    BIND = 'bind'
    UNBIND = "unbind"
    UNBINDALL = "unbindall"

class VXLANHandler(RoutineContainer):
    def __init__(self,connection,parent):
        super(VXLANHandler,self).__init__(connection.scheduler)
        self._conn = connection
        self._parent = parent

        self.vxlan_vlan_map_info = {}
        self.vlanid_pool = parent.vlanid_pool
        self._last_lognets_info = {}
        self._last_phyport_info = {}

        self._store_event = dict()
        self._store_bind_event = []
        self._store_unbind_event = []
        self._store_unbindall_event = []

        if self._parent.remote_api:
            self.api = remoteAPI
        else:
            self.api = callAPI

    def main(self):

        try:
            data_change_event = ioprocessing.DataObjectChanged.createMatcher(None, None, self._conn)

            # routine sync call to vtep controller (updatelogicalswitch,unbindlogicalswitch,unbindphysicalport)
            self.subroutine(self.action_handler(),True,"_action_handler")

            _last_change_event = [None]
            while True:
                yield (data_change_event,)
                _last_change_event[0] = self.event
                def _update_callback(event, matcher):
                    _last_change_event[0] = event
                def _update_loop():
                    while _last_change_event[0] is not None:
                        try:
                            last_lognets_info = self._last_lognets_info
                            last_phyport_info = self._last_phyport_info
        
                            current_logports,current_phyports,current_lognets,_ = _last_change_event[0].current
                            _last_change_event[0] = None
        
                            lognet_to_logports = {}
                            for p, _ in current_logports:
                                lognet_to_logports.setdefault(p.network, []).append(p.id)
        
                            datapath_id = self._conn.openflow_datapathid
                            vhost = self._conn.protocol.vhost
        
                            def get_phyport_info(portid):
                                try:
                                    for m in callAPI(self, "ovsdbportmanager", "waitportbyno",
                                                 {"datapathid": datapath_id,
                                                  "vhost": vhost,
                                                  "timeout": 1,
                                                  "portno": portid}):
                                        yield m
                                except ovsdbportmanager.OVSDBPortNotAppearException:
                                    self.retvalue = None
                            phy_ports = [p for p in current_phyports if _is_vxlan(p[0])]
                            for m in self.executeAll([get_phyport_info(pid) for p,pid in phy_ports]):
                                yield m
        
                            phy_port_info  = dict((k[0], (v[0]['external_ids']['vtep-physname'],
                                                                v[0]['external_ids']['vtep-phyiname']))
                                                  for k,v in zip(phy_ports,self.retvalue)
                                                  if v[0] is not None and \
                                                    'vtep-physname' in v[0]['external_ids'] and \
                                                    'vtep-phyiname' in v[0]['external_ids'])
                            unique_phyports = dict((p.physicalnetwork, p)
                                                   for p,_ in sorted(phy_ports, key = lambda x: x[1])
                                                   if p in phy_port_info)
        
                            lognet_to_phyport =  dict((n,unique_phyports[n.physicalnetwork])
                                                      for n,_ in current_lognets
                                                      if n.physicalnetwork in unique_phyports)
        
                            current_lognets_info = dict((n,(phy_port_info[v], lognet_to_logports[n]))
                                                        for n,v in lognet_to_phyport.items()
                                                        if n in lognet_to_logports)
        
                            self._last_lognets_info = current_lognets_info
                            self._last_phyport_info = phy_port_info
        
                            # add or update new phyport, unbind all physical interface bind
                            # must do it as init first ....
                            for p in phy_port_info:
                                if p not in last_phyport_info or\
                                        (p in last_phyport_info and last_phyport_info[p] != phy_port_info[p]):
        
                                    physname = phy_port_info[p][0]
                                    phyiname = phy_port_info[p][1]
        
                                    event = VtepControllerCall(connection=self._conn,type=VtepControllerCall.UNBINDALL,
                                                               physname=physname,phyiname=phyiname)
        
                                    for m in self.waitForSend(event):
                                        yield m
        
        
                            # remove phyport , unbind all physical interface bind also
                            for p in last_phyport_info:
                                if p not in phy_port_info or\
                                        (p in phy_port_info and last_phyport_info[p] != phy_port_info[p]):
                                    physname = last_phyport_info[p][0]
                                    phyiname = last_phyport_info[p][1]
        
                                    event = VtepControllerCall(connection=self._conn,type=VtepControllerCall.UNBINDALL,
                                                               physname=physname,phyiname=phyiname)
        
                                    for m in self.waitForSend(event):
                                        yield m
        
                            for n in last_lognets_info:
                                if n not in current_lognets_info:
                                    # means lognet removed
                                    # release vlan id
                                    # unbind physwitch
        
                                    # if last cause exception , n.id mybe not in vxlan_vlan_map_info
                                    if n.id in self.vxlan_vlan_map_info:
                                        vlan_id = self.vxlan_vlan_map_info[n.id]
                                        physname = last_lognets_info[n][0][0]
                                        phyiname = last_lognets_info[n][0][1]
        
                                        del self.vxlan_vlan_map_info[n.id]
        
                                        event = VXLANMapChanged(self._conn,n.id,VXLANMapChanged.DELETED)
        
                                        for m in self.waitForSend(event):
                                            yield m
        
                                        event = VtepControllerCall(connection=self._conn,logicalnetworkid=n.id,
                                                                   physname=physname,phyiname=phyiname,vlanid=vlan_id,
                                                                   type=VtepControllerCall.UNBIND)
        
                                        for m in self.waitForSend(event):
                                            yield m
                                
                                else:
                                    last_info = last_lognets_info[n]
                                    current_info = current_lognets_info[n]
        
                                    if last_info[0] != current_info[0]:
                                        # physwitch info have changed
                                        # unbind it
        
                                        if n.id in self.vxlan_vlan_map_info:
                                            vid = self.vxlan_vlan_map_info[n.id]
                                            physname = last_lognets_info[n][0][0]
                                            phyiname = last_lognets_info[n][0][1]
        
                                            del self.vxlan_vlan_map_info[n.id]
        
                                            event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
                                                                       physname=physname, phyiname=phyiname, vlanid=vid,
                                                                       type=VtepControllerCall.UNBIND)
        
                                            for m in self.waitForSend(event):
                                                yield m
                                           
                                    else:
                                        if last_info[1] != current_info[1]:
                                            add_ports = list(set(current_info[1]).difference(set(last_info[1])))
        
                                            if add_ports:
                                                # add new add_ports
                                                # update bind logical ports
        
                                                if n.id in self.vxlan_vlan_map_info:
                                                    vid = self.vxlan_vlan_map_info[n.id]
                                                    physname = last_lognets_info[n][0][0]
                                                    phyiname = last_lognets_info[n][0][1]
        
                                                    event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
                                                                               vni = n.vni,logicalports=add_ports,
                                                                               physname=physname, phyiname=phyiname, vlanid=vid,
                                                                               type=VtepControllerCall.BIND)
        
                                                    for m in self.waitForSend(event):
                                                        yield m
        
        
                            for n in current_lognets_info:
                                if n not in last_lognets_info:
                                    # means add lognet
                                    # find avaliable vlan id , send event
                                    # bind physwitch
                                    find = False
                                    _current_vlan_ids = set(self.vxlan_vlan_map_info.values())
                                    for start,end in self.vlanid_pool:
                                        for vid in range(start,end + 1):
                                            if vid not in _current_vlan_ids:
                                                find = True
                                                break
                                        if find:
                                            break
        
                                    if find:
                                        self.vxlan_vlan_map_info[n.id] = vid
        
                                        event = VXLANMapChanged(self._conn,n.id,VXLANMapChanged.UPDATED,vlan_id = vid)
                                        for m in self.waitForSend(event):
                                            yield m
        
                                        physname = current_lognets_info[n][0][0]
                                        phyiname = current_lognets_info[n][0][1]
                                        ports = current_lognets_info[n][1]
        
                                        event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
                                                                   vni=n.vni, logicalports=ports,
                                                                   physname=physname, phyiname=phyiname, vlanid=vid,
                                                                   type=VtepControllerCall.BIND)
        
                                        for m in self.waitForSend(event):
                                            yield m
        
                                    else:
                                        self._parent._logger.warning('Not enough vlan_id for logical network %r', n.id)
                                        event = VXLANMapChanged(self._conn,n.id,VXLANMapChanged.UPDATED,vlan_id = None)
                                        for m in self.waitForSend(event):
                                            yield m
                                else:
        
                                    last_info = last_lognets_info[n]
                                    current_info = current_lognets_info[n]
        
                                    if last_info[0] != current_info[0]:
                                        # physwitch info have changed
                                        # rebind it
                                        if n.id in self.vxlan_vlan_map_info:
                                            vid = self.vxlan_vlan_map_info[n.id]
                                            physname = current_lognets_info[n][0][0]
                                            phyiname = current_lognets_info[n][0][1]
                                            ports = current_lognets_info[n][1]
        
                                            event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
                                                                       vni=n.vni, logicalports=ports,
                                                                       physname=physname, phyiname=phyiname, vlanid=vid,
                                                                       type=VtepControllerCall.BIND)
        
                                            for m in self.waitForSend(event):
                                                yield m
        
        
                        except Exception:
                            self._parent._logger.info(" vxlan vtep handler exception , continue", exc_info = True)
                for m in self.withCallback(_update_loop(), _update_callback, data_change_event):
                    yield m
        finally:

            if hasattr(self,"_action_handler"):
                self._action_handler.close()


# ===============================================================================
#     def time_cycle(self):
#
#         while True:
#             for m in self.waitWithTimeout(self._parent.refreshinterval):
#                 yield m
#
#             for n,v in self._last_lognets_info.items():
#
#                 if n.id in self.vxlan_vlan_map_info:
#                     vid = self.vxlan_vlan_map_info[n.id]
#                     physname = v[0][0]
#                     phyiname = v[0][1]
#                     ports = v[1]
#
#                     event = VtepControllerCall(connection=self._conn, logicalnetworkid=n.id,
#                                                                vni=n.vni, logicalports=ports,
#                                                                physname=physname, phyiname=phyiname, vlanid=vid,
#                                                                type=VtepControllerCall.BIND)
#                     self.scheduler.emergesend(event)
# ===============================================================================

    def action_handler(self):

        bind_event = VtepControllerCall.createMatcher(self._conn)
        event_queue = []
        timeout_flag = [False]

        def handle_action():
            while event_queue or timeout_flag[0]:
                events = event_queue[:]
                del event_queue[:]

                for e in events:
                    # every event must have physname , phyiname
                    physname = e.physname
                    phyiname = e.phyiname
                    if e.type == VtepControllerCall.UNBINDALL:
                        # clear all other event info
                        self._store_event[(physname,phyiname)] = {"all":e}
                    elif e.type == VtepControllerCall.BIND:
                        # bind will combine bind event before
                        vlanid = e.vlanid
                        if (physname,phyiname) in self._store_event:
                            v = self._store_event[(physname,phyiname)]

                            if vlanid in v:
                                x = v[vlanid]
                                logicalports = e.logicalports

                                if x[0] == VtepControllerCall.BIND and x[1] == e.logicalnetworkid:
                                    new_set = set(e.logicalports)
                                    old_set = set(x[3])
                                    new_set.update(old_set)
                                    logicalports = list(new_set)

                                v.update({vlanid:(e.type,e.logicalnetworkid,e.vni,logicalports)})
                                self._store_event[(physname,phyiname)] = v
                            else:
                                # new bind info , no combind event
                                v.update({vlanid:(e.type,e.logicalnetworkid,e.vni,e.logicalports)})
                                self._store_event[(physname,phyiname)] = v
                        else:
                            self._store_event[(physname,phyiname)] = {vlanid:(e.type,e.logicalnetworkid,
                                                                              e.vni,e.logicalports)}

                    elif e.type == VtepControllerCall.UNBIND:

                        vlanid = e.vlanid

                        if (physname,phyiname) in self._store_event:
                            v = self._store_event[(physname,phyiname)]
                            v.update({vlanid:(e.type,e.logicalnetworkid)})
                            self._store_event[(physname,phyiname)] = v
                        else:
                            self._store_event[(physname,phyiname)] = {vlanid:(e.type,e.logicalnetworkid)}

                    else:
                        self._parent._logger.warning("catch error type event %r , ignore it", e)
                        continue

                call = []
                target_name = "vtepcontroller"
                for k,v in self._store_event.items():
                    if "all" in v:
                        # send unbindall
                        call.append(self.api(self,target_name,"unbindphysicalport",
                                             {"physicalswitch": k[0], "physicalport": k[1]},
                                             timeout=10))
                        # unbindall , del it whatever
                        del v["all"]

                try:
                    for m in self.executeAll(call):
                        yield m
                except Exception:
                    self._parent._logger.warning("unbindall remove call failed")

                for k,v in self._store_event.items():
                    for vlanid , e in dict(v).items():
                        if vlanid != "all":
                            if e[0] == VtepControllerCall.BIND:

                                params = {"physicalswitch": k[0],
                                            "physicalport": k[1],
                                            "vlanid": vlanid,
                                            "logicalnetwork": e[1],
                                            "vni":e[2],
                                            "logicalports": e[3]}

                                try:
                                    for m in self.api(self,target_name,"updatelogicalswitch",
                                                  params,timeout=10):
                                        yield m
                                except Exception:
                                    self._parent._logger.warning("update logical switch error,try next %r",params)
                                else:
                                    del self._store_event[k][vlanid]

                            if e[0] == VtepControllerCall.UNBIND:

                                params = {"logicalnetwork":e[1],
                                                "physicalswitch":k[0],
                                                "physicalport":k[1],
                                                  "vlanid":vlanid}

                                try:
                                    for m in self.api(self,target_name,"unbindlogicalswitch",
                                                      params,timeout=10):
                                        yield m
                                except Exception:
                                    self._parent._logger.warning("unbind logical switch error,try next %r",params)
                                else:
                                    del self._store_event[k][vlanid]

                self._store_event = dict((k,v) for k,v in self._store_event.items() if v )

                if timeout_flag[0]:
                    timeout_flag[0] = False

        def append_event(event,matcher):
            event_queue.append(event)

        while True:

            for m in self.waitWithTimeout(10,bind_event):
                yield m

            if not self.timeout:
                event_queue.append(self.event)
            else:
                timeout_flag[0] = True

            for m in self.withCallback(handle_action(),append_event,bind_event):
                yield m


    def wait_vxlan_map_info(self,container,logicalnetworkid,timeout = 4):

        if logicalnetworkid in self.vxlan_vlan_map_info:
            container.retvalue = self.vxlan_vlan_map_info[logicalnetworkid]
        else:
            event = VXLANMapChanged.createMatcher(self._conn,logicalnetworkid,VXLANMapChanged.UPDATED)

            for m in container.waitWithTimeout(timeout,event):
                yield m

            if container.timeout:
                raise ValueError(" cannot find vxlan vlan map info ")
            else:
                container.retvalue = container.event.vlan_id

@defaultconfig
class VXLANVtep(FlowBase):
    """
    Use hardware_vtep instead of software VXLAN
    """
    # Use these VLANs for vtep configuration. Must not be conflicted with VLAN networks.
    _default_vlanid_pool = ((3000, 4000),)

    # remote_api means call remote api to vtep controller
    _default_remote_api = True

    # interval time to retry failed actions
    _default_retryactioninterval = 60

    def __init__(self,server):
        super(VXLANVtep,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)

        self.conns = {}
        self.vxlan_vlan_map_info = {}

        self.createAPI(publicapi(self.createioflowparts, self.app_routine,
                                 lambda connection,logicalnetwork,**kwargs:
                                        _is_vxlan(logicalnetwork)),
                       api(self.get_vxlan_bind_info,self.app_routine))

    def _main(self):

        # check vlan pool
        lastend = 0
        for start,end in self.vlanid_pool:
            if start > end or start < lastend:
                raise ValueError(" vlan sequences overlapped or disorder ")
            if end > 4095:
                raise ValueError(" vlan out of range (0 -- 4095) ")
            lastend = end

        flow_init = FlowInitialize.createMatcher(_ismatch= lambda x : self.vhostbind is None
                                                 or x.vhost in self.vhostbind)
        conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                               _ismatch= lambda x:self.vhostbind is None
                                                               or x.createby.vhost in self.vhostbind)

        while True:
            yield (flow_init,conn_down)

            if self.app_routine.matcher is flow_init:

                conn = self.app_routine.event.connection

                self.app_routine.subroutine(self._init_conn(conn))

            if self.app_routine.matcher is conn_down:
                conn = self.app_routine.event.connection

                self.app_routine.subroutine(self._uninit_conn(conn))

    def _init_conn(self,conn):
        if conn in self.conns:
            handler = self.conns.pop(conn)
            handler.close()

        handler = VXLANHandler(conn,self)
        handler.start()
        self.conns[conn] = handler

        if None:
            yield

    def _uninit_conn(self,conn):
        if conn in self.conns:
            handler = self.conns.pop(conn)
            handler.close()

        if None:
            yield

    def get_vxlan_bind_info(self,systemid=None):
        """get vxlan -> vlan , bind info"""

        ret = []
        for conn in self.conns:
            datapath_id = conn.openflow_datapathid
            vhost = conn.protocol.vhost

            for m in callAPI(self.app_routine,"ovsdbmanager","getbridgeinfo",{"datapathid":datapath_id,
                                                                              "vhost":vhost}):
                yield m

            if self.app_routine.retvalue:
                _,system_id,_ = self.app_routine.retvalue

                if systemid:
                    if systemid == system_id:
                        handler = self.conns[conn]
                        ret.append({system_id: handler.vxlan_vlan_map_info})
                        break
                else:
                    handler = self.conns[conn]
                    ret.append({system_id:handler.vxlan_vlan_map_info})


        self.app_routine.retvalue = ret

    def createioflowparts(self,connection,logicalnetwork,physicalport,logicalnetworkid,physicalportid):

        self._logger.debug(" create flow parts connection = %r",connection)
        self._logger.debug(" create flow parts logicalnetworkid = %r", logicalnetworkid)
        self._logger.debug(" create flow parts physicalportid = %r", physicalportid)

        find = False
        vid = None
        if connection in self.conns:

            handler = self.conns[connection]

            try:
                for m in handler.wait_vxlan_map_info(self.app_routine,logicalnetwork.id):
                    yield m
            except Exception:
                find = False
                self._logger.warning(" get vxlan vlan map info error", exc_info = True)
            else:
                find = True
                vid = self.app_routine.retvalue
                if vid is None:
                    self._logger.warning('Not enough vlan ID, io flow parts not created')
                    find = False

        if find:
            self.app_routine.retvalue = self.createflowparts(connection, vid, physicalportid)
        else:
            self.app_routine.retvalue = ([],[],[],[],[])

    def createflowparts(self,connection,vid,physicalportid):
        ofdef = connection.openflowdef
        in_flow_match_part = [
            ofdef.create_oxm(ofdef.OXM_OF_VLAN_VID, vid | ofdef.OFPVID_PRESENT)
        ]

        in_flow_action_part = [
            ofdef.ofp_action(type=ofdef.OFPAT_POP_VLAN)
        ]

        out_flow_action_part = [
            ofdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
            ofdef.ofp_action_set_field(
                field=ofdef.create_oxm(
                    ofdef.OXM_OF_VLAN_VID,
                    vid | ofdef.OFPVID_PRESENT
                )
            ),
            ofdef.ofp_action_output(
                port=physicalportid
            )
        ]

        out_flow_action_part2 = [
            ofdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
            ofdef.ofp_action_set_field(
                field=ofdef.create_oxm(
                    ofdef.OXM_OF_VLAN_VID,
                    vid | ofdef.OFPVID_PRESENT
                )
            ),
            ofdef.ofp_action_output(
                port=ofdef.OFPP_IN_PORT
            )
        ]

        group_buckets = [
            ofdef.ofp_action_push(ethertype=ETHERTYPE_8021Q),
            ofdef.ofp_action_set_field(
                field=ofdef.create_oxm(
                    ofdef.OXM_OF_VLAN_VID,
                    vid | ofdef.OFPVID_PRESENT
                )
            ),
            ofdef.ofp_action_output(
                port=physicalportid
            )
        ]

        return in_flow_match_part, in_flow_action_part,out_flow_action_part, group_buckets,out_flow_action_part2



'''
Created on 2015/9/30/

:author: hubo
'''
from vlcp.config import Configurable, defaultconfig
from vlcp.event import Event, withIndices
import logging
from vlcp.event.runnable import RoutineContainer, EventHandler
import sys
from vlcp.utils.pycache import removeCache
import functools
import copy
from vlcp.config.config import manager
from vlcp.event.core import QuitException
from inspect import cleandoc

try:
    reload
except:
    try:
        from importlib import reload
    except:
        from imp import reload


def depend(*args):
    """
    Decorator to declare dependencies to other modules. Recommended usage is::
    
        import other_module
        
        @depend(other_module.ModuleClass)
        class MyModule(Module):
            ...
    
    :param \*args: depended module classes.
    """
    def decfunc(cls):
        if not 'depends' in cls.__dict__:
            cls.depends = []
        cls.depends.extend(list(args))
        for a in args:
            if not hasattr(a, 'referencedBy'):
                a.referencedBy = []
            a.referencedBy.append(cls)
        return cls
    return decfunc

@withIndices('target', 'name')
class ModuleNotification(Event):
    pass

@withIndices('handle', 'target', 'name')
class ModuleAPICall(Event):
    canignore = False

@withIndices('handle')
class ModuleAPIReply(Event):
    pass

class ModuleAPICallTimeoutException(Exception):
    pass

@withIndices('module', 'state', 'instance')
class ModuleLoadStateChanged(Event):
    NOTLOADED = 'notloaded'
    LOADING = 'loading'
    LOADED = 'loaded'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'
    UNLOADING = 'unloading'
    UNLOADED = 'unloaded'

def create_discover_info(func):
    code = func.__code__
    if code.co_flags & 0x08:
        haskwargs = True
    else:
        haskwargs = False
    # Remove argument env
    arguments = code.co_varnames[0:code.co_argcount]
    if hasattr(func, '__self__') and func.__self__:
        # First argument is self, remove an extra argument
        arguments=arguments[1:]
    # Optional arguments
    if hasattr(func, '__defaults__') and func.__defaults__:
        requires = arguments[:-len(func.__defaults__)]
        optionals = arguments[-len(func.__defaults__):]
    else:
        requires = arguments[:]
        optionals = []
    return {'description': cleandoc(func.__doc__) if func.__doc__ is not None else '',
            'parameters':
                [{'name':n,'optional':False} for n in requires]
                + [{'name':optionals[i],'optional':True,'default':func.__defaults__[i]}
                   for i in range(0,len(optionals))],
            'extraparameters': haskwargs
                }
    

def api(func, container = None, criteria = None):
    '''
    Return an API def for a generic function
    
    :param func: a function or bounded method
    
    :param container: if None, this is used as a synchronous method, the return value of the method
                      is used for the return value. If not None, this is used as an asynchronous method,
                      the return value should be a generator, and it is executed in `container` as a routine.
                      The return value should be set to `container.retvalue`.
    
    :param criteria: An extra function used to test whether this function should process the API. This allows
                     multiple API definitions to use the same API method name.
    '''
    return (func.__name__.lower(), functools.update_wrapper(lambda n,p: func(**p), func), container,
            create_discover_info(func), criteria)

def publicapi(func, container = None, criteria = None):
    '''
    Create an API def for public API processing. Target name of a public API is `'public'`.

    :param func: a function or bounded method
    
    :param container: if None, this is used as a synchronous method, the return value of the method
                      is used for the return value. If not None, this is used as an asynchronous method,
                      the return value should be a generator, and it is executed in `container` as a routine.
                      The return value should be set to `container.retvalue`.
    
    :param criteria: An extra function used to test whether this function should process the API. This allows
                     multiple API definitions to use the same API method name.
    '''
    return ("public/" + func.__name__.lower(), functools.update_wrapper(lambda n,p: func(**p), func), container,
            create_discover_info(func), criteria)

class ModuleAPIHandler(RoutineContainer):
    """
    API Handler for modules
    """
    def __init__(self, moduleinst, apidefs = None, allowdiscover = True, rejectunknown = True):
        RoutineContainer.__init__(self, scheduler=moduleinst.scheduler, daemon=False)
        self.handler = EventHandler(self.scheduler)
        self.servicename = moduleinst.getServiceName()
        self.apidefs = apidefs
        self.registeredAPIs = {}
        self.discoverinfo = {}
        self.allowdiscover = allowdiscover
        self.rejectunknown = True
    @staticmethod
    def createReply(handle, result):
        return ModuleAPIReply(handle, result=result)
    @staticmethod
    def createExceptionReply(handle, exception):
        return ModuleAPIReply(handle, exception = exception)
    def _createHandler(self, name, handler, container = None, discoverinfo = None, criteria = None):
        extra_params = {}
        if criteria:
            extra_params['_ismatch'] = lambda e: not e.canignore and criteria(**e.params)
        if name is None:
            matcher = ModuleAPICall.createMatcher(target = self.servicename, **extra_params)
            matcher = ModuleAPICall.createMatcher(target = self.servicename, **extra_params)
        elif name.startswith('public/'):
            matcher = ModuleAPICall.createMatcher(target = 'public', name = name[len('public/'):], **extra_params)
        else:
            matcher = ModuleAPICall.createMatcher(target = self.servicename, name = name, **extra_params)
        if container is not None:
            def wrapper(event):
                try:
                    for m in handler(event.name, event.params):
                        yield m
                    for m in container.waitForSend(self.createReply(event.handle, container.retvalue)):
                        yield m
                except Exception as val:
                    for m in container.waitForSend(self.createExceptionReply(event.handle, val)):
                        yield m
            def event_handler(event, scheduler):
                event.canignore = True
                container.subroutine(wrapper(event), False)
        else:
            def wrapper(event):
                try:
                    result = handler(event.name, event.params)
                    for m in self.waitForSend(self.createReply(event.handle, result)):
                        yield m
                except Exception as val:
                    for m in self.waitForSend(self.createExceptionReply(event.handle, val)):
                        yield m
            def event_handler(event, scheduler):
                event.canignore = True
                self.subroutine(wrapper(event), False)
        return (matcher, event_handler)
    def registerAPIs(self, apidefs):
        '''
        API definition is in format: `(name, handler, container, discoverinfo)`
        
        if the handler is a generator, container should be specified
        handler should accept two arguments::
        
            def handler(name, params):
                ...
        
        `name` is the method name, `params` is a dictionary contains the parameters.
        
        the handler can either return the result directly, or be a generator (async-api),
        and write the result to container.retvalue on exit.
        e.g::
        
            ('method1', self.method1),    # method1 directly returns the result
            ('method2', self.method2, self) # method2 is an async-api
        
        Use api() to automatically generate API definitions.
        '''
        handlers = [self._createHandler(*apidef) for apidef in apidefs]
        self.handler.registerAllHandlers(handlers)
        self.discoverinfo.update((apidef[0], apidef[3] if len(apidef) > 3 else {'description':cleandoc(apidef[1].__doc__)}) for apidef in apidefs)
    def registerAPI(self, name, handler, container = None, discoverinfo = None, criteria = None):
        """
        Append new API to this handler
        """
        self.handler.registerHandler(*self._createHandler(name, handler, container, criteria))
        if discoverinfo is None:
            self.discoverinfo[name] = {'description': cleandoc(handler.__doc__)}
        else:
            self.discoverinfo[name] = discoverinfo
    def unregisterAPI(self, name):
        """
        Remove an API from this handler
        """
        if name.startswith('public/'):
            target = 'public'
            name = name[len('public/'):]
        else:
            target = self.servicename
            name = name
        removes = [m for m in self.handler.handlers.keys() if m.target == target and m.name == name]
        for m in removes:
            self.handler.unregisterHandler(m)
    def discover(self, details = False):
        'Discover API definitions. Set details=true to show details'
        if details and not (isinstance(details, str) and details.lower() == 'false'):
            return copy.deepcopy(self.discoverinfo)
        else:
            return dict((k,v.get('description', '')) for k,v in self.discoverinfo.items())
    def reject(self, name, args):
        raise ValueError('%r is not defined in module %r' % (name, self.servicename))
    def start(self, asyncStart=False):
        if self.apidefs:
            self.registerAPIs(self.apidefs)
        if self.allowdiscover:
            self.registerAPI(*api(self.discover))
        if self.rejectunknown:
            self.handler.registerHandler(*self._createHandler(None, self.reject, None))
    def close(self):
        self.handler.close()

@defaultconfig
class Module(Configurable):
    '''
    A functional part which can be loaded or unloaded dynamically
    '''
    # Service modules are automatically unloaded if all the dependencies are unloaded
    _default_service = False
    # Force stop all the RoutineContainers that are started by this module.
    # Subroutines are not stopped.
    _default_forcestop = True
    # Automatically change module status to SUCCEEDED when load() is finished
    _default_autosuccess = True
    depends = []
    def __init__(self, server):
        '''
        Constructor
        '''
        Configurable.__init__(self)
        self.server = server
        self.scheduler = server.scheduler
        self.connections = []
        self.routines = []
        self.dependedBy = set()
        self.state = ModuleLoadStateChanged.NOTLOADED
        self.target = type(self)
        self._logger = logging.getLogger(type(self).__module__ + '.' + type(self).__name__)
    @classmethod
    def getFullPath(cls):
        return cls.__module__ + '.' + cls.__name__
    def createAPI(self, *apidefs):
        """
        Create API definitions on this module. This creates a ModuleAPIHandler and register
        these `apidefs` to it.
        
        :param \*apidefs: should be return values of `api()` or `publicapi()` functions.
        """
        if hasattr(self, 'apiHandler'):
            self.appendAPI(*apidefs)
        self.apiHandler = ModuleAPIHandler(self, apidefs)
        self.routines.append(self.apiHandler)
    def appendAPI(self, *apidefs):
        t = list(self.apiHandler.apidefs)
        t.extend(apidefs)
        self.apiHandler.apidefs = t
    def getServiceName(self):
        """
        Return the targetname (or servicename) for this module
        """
        if hasattr(self, 'servicename'):
            return self.servicename
        else:
            return self.target.__name__.lower()
    def load(self, container):
        '''
        Load module
        '''
        self.target._instance = self
        try:
            for r in self.routines:
                r.start()
            for c in self.connections:
                c.start()
            try:
                for m in self.changestate(ModuleLoadStateChanged.LOADED, container):
                    yield m
                if self.autosuccess:
                    for m in self.changestate(ModuleLoadStateChanged.SUCCEEDED, container):
                        yield m
            except ValueError:
                pass
        except:
            for m in self.changestate(ModuleLoadStateChanged.FAILED, container):
                yield m
            raise
    def unload(self, container, force = False):
        '''
        Unload module
        '''
        for m in self.changestate(ModuleLoadStateChanged.UNLOADING, container):
            yield m
        for c in self.connections:
            try:
                for m in c.shutdown():
                    yield m
            except:
                self._logger.warning('Except when shutting down connection %r', c, exc_info = True)
        if self.forcestop or force:
            for r in self.routines:
                try:
                    r.close()
                except:
                    self._logger.warning('Except when unloading module', exc_info = True)
        for m in self.changestate(ModuleLoadStateChanged.UNLOADED, container):
            yield m
        if hasattr(self.target, '_instance') and self.target._instance is self:
            del self.target._instance
    _changeMap = set(((ModuleLoadStateChanged.NOTLOADED, ModuleLoadStateChanged.LOADING),
                 (ModuleLoadStateChanged.LOADING, ModuleLoadStateChanged.LOADED),
                 (ModuleLoadStateChanged.LOADING, ModuleLoadStateChanged.SUCCEEDED),
                 (ModuleLoadStateChanged.LOADED, ModuleLoadStateChanged.SUCCEEDED),
                 (ModuleLoadStateChanged.LOADING, ModuleLoadStateChanged.FAILED),
                 (ModuleLoadStateChanged.LOADED, ModuleLoadStateChanged.FAILED),
                 (ModuleLoadStateChanged.SUCCEEDED, ModuleLoadStateChanged.UNLOADING),
                 (ModuleLoadStateChanged.FAILED, ModuleLoadStateChanged.UNLOADING),
                 (ModuleLoadStateChanged.UNLOADING, ModuleLoadStateChanged.UNLOADED)))
    def changestate(self, state, container):
        """
        Change the current load state.
        """
        if self.state != state:
            if not (self.state, state) in self._changeMap:
                raise ValueError('Cannot change state from %r to %r' % (self.state, state))
            self.state = state
            self._logger.debug('Module state changed to %r', state)
            for m in container.waitForSend(ModuleLoadStateChanged(self.target, state, self)):
                yield m


class ModuleLoadException(Exception):
    """
    Raised when module loading failed.
    """
    pass

def findModule(path, autoimport = True):
    dotpos = path.rfind('.')
    if dotpos == -1:
        raise ModuleLoadException('Must specify module with full path, including package name')
    package = path[:dotpos]
    classname = path[dotpos+1:]
    p = None
    module = None
    try:
        p = sys.modules[package]
        module = getattr(p, classname)
    except KeyError:
        pass
    except AttributeError:
        pass
    if p is None and autoimport:
        p = __import__(package, fromlist=(classname,))
        module = getattr(p, classname)
    return (p, module)

class _ProxyMetaClass(type):
    '''
    Metaclass for delayed depend
    '''
    @property
    def depends(self):
        if not hasattr(self, '_depends'):
            self._depends = []
            path = manager.get('proxy.' + self.__name__.lower())
            if path is not None:
                try:
                    _, module = findModule(path, True)
                except KeyError as exc:
                    raise ModuleLoadException('Cannot load module ' + repr(path) + ': ' + str(exc) + 'is not defined in the package')
                except Exception as exc:
                    raise ModuleLoadException('Cannot load module ' + repr(path) + ': ' + str(exc))
                if module is None:
                    raise ModuleLoadException('Cannot find module: ' + repr(path))
            else:
                module = self._default
                if module is None:
                    raise ModuleLoadException('Cannot find module: proxy.' + self.__name__ + ' is not specified')
            self._proxytarget = module
            self.depends.append(module)
            if not hasattr(module, 'referencedBy'):
                module.referencedBy = []
            module.referencedBy.append(self)
        return self._depends
        
class _ProxyModule(Module):
    '''
    A proxy to create dependencies on configurable module
    '''
    service = True
    def __init__(self, server):
        '''
        Constructor
        '''
        Module.__init__(self, server)        
        self.proxyhandler = EventHandler(self.scheduler)
    def load(self, container):
        self._targetname = self._proxytarget._instance.getServiceName()
        self.proxyhandler.registerHandler(ModuleAPICall.createMatcher(None, self.getServiceName()), self._proxyhandler)
        for m in Module.load(self, container):
            yield m
    def unload(self, container, force=False):
        self.proxyhandler.close()
        for m in Module.unload(self, container, force=force):
            yield m
    def _proxyhandler(self, event, scheduler):
        event.canignore = True
        scheduler.emergesend(ModuleAPICall(event.handle, self._targetname, event.name, params = event.params))        

def proxy(name, default = None):
    """
    Create a proxy module. A proxy module has a default implementation, but can be redirected to other
    implementations with configurations. Other modules can depend on proxy modules.
    """
    proxymodule = _ProxyMetaClass(name, (_ProxyModule,), {'_default': default})
    proxymodule.__module__ = sys._getframe(1).f_globals.get('__name__')
    return proxymodule

class ModuleLoader(RoutineContainer):
    """
    Module loader to load modules. The server object creates this instance automatically, usually
    you can retrieve the pre-created object from `server.moduleloader`
    """
    _logger = logging.getLogger(__name__ + '.ModuleLoader')
    def __init__(self, server):
        self.server = server
        RoutineContainer.__init__(self, scheduler=server.scheduler, daemon=True)
        self.activeModules = {}
    def main(self):
        try:
            # Reject unmatched public API
            public_api = ModuleAPICall.createMatcher(target = "public")
            while True:
                yield (public_api,)
                if not self.event.canignore:
                    self.event.canignore = True
                    self.scheduler.emergesend(ModuleAPIHandler.createExceptionReply(self.event.handle, ValueError("public API is not processed")))
        except QuitException:
            c = ModuleAPICall.createMatcher()
            while True:
                yield (c,)
                if not self.event.canignore:
                    self.event.canignore = True
                    self.scheduler.emergesend(ModuleAPIHandler.createExceptionReply(self.event.handle, QuitException()))
    def _removeDepend(self, module, depend):
        depend._instance.dependedBy.remove(module)
        if not depend._instance.dependedBy and depend._instance.service:
            # Automatically unload a service which is not used
            self.subroutine(self.unloadmodule(depend), False)
    def loadmodule(self, module):
        '''
        Load a module class. Need delegate to call from other containers.
        '''
        self._logger.debug('Try to load module %r', module)
        if hasattr(module, '_instance'):
            self._logger.debug('Module is already initialized, module state is: %r', module._instance.state)
            if module._instance.state == ModuleLoadStateChanged.UNLOADING or module._instance.state == ModuleLoadStateChanged.UNLOADED:
                # Wait for unload
                um = ModuleLoadStateChanged.createMatcher(module._instance.target, ModuleLoadStateChanged.UNLOADED)
                yield (um,)
            elif module._instance.state == ModuleLoadStateChanged.SUCCEEDED:
                pass
            elif module._instance.state == ModuleLoadStateChanged.FAILED:
                raise ModuleLoadException('Cannot load module %r before unloading the failed instance' % (module,))
            elif module._instance.state == ModuleLoadStateChanged.NOTLOADED or module._instance.state == ModuleLoadStateChanged.LOADED or module._instance.state == ModuleLoadStateChanged.LOADING:
                # Wait for succeeded or failed
                sm = ModuleLoadStateChanged.createMatcher(module._instance.target, ModuleLoadStateChanged.SUCCEEDED)
                fm = ModuleLoadStateChanged.createMatcher(module._instance.target, ModuleLoadStateChanged.FAILED)
                yield (sm, fm)
                if self.matcher is sm:
                    pass
                else:
                    raise ModuleLoadException('Module load failed for %r' % (module,))
        if not hasattr(module, '_instance'):
            self._logger.info('Loading module %r', module)
            inst = module(self.server)
            # Avoid duplicated loading
            module._instance = inst
            # When reloading, some of the dependencies are broken, repair them
            if hasattr(module, 'referencedBy'):
                inst.dependedBy = set([m for m in module.referencedBy if hasattr(m, '_instance') and m._instance.state != ModuleLoadStateChanged.UNLOADED])
            # Load Module
            for d in module.depends:
                if hasattr(d, '_instance') and d._instance.state == ModuleLoadStateChanged.FAILED:
                    raise ModuleLoadException('Cannot load module %r, it depends on %r, which is in failed state.' % (module, d))
            try:
                for d in module.depends:
                    if hasattr(d, '_instance') and d._instance.state == ModuleLoadStateChanged.SUCCEEDED:
                        d._instance.dependedBy.add(module)
                preloads = [d for d in module.depends if not hasattr(d, '_instance') or \
                            d._instance.state != ModuleLoadStateChanged.SUCCEEDED]
                for d in preloads:
                    self.subroutine(self.loadmodule(d), False)
                sms = [ModuleLoadStateChanged.createMatcher(d, ModuleLoadStateChanged.SUCCEEDED) for d in preloads]
                fms = [ModuleLoadStateChanged.createMatcher(d, ModuleLoadStateChanged.FAILED) for d in preloads]
                ms = sms + fms
                while sms:
                    yield ms
                    if self.matcher in fms:
                        raise ModuleLoadException('Cannot load module %r, it depends on %r, which is in failed state.' % (module, self.event.module))
                    else:
                        sms.remove(self.matcher)
                        ms.remove(self.matcher)
            except:
                for d in module.depends:
                    if hasattr(d, '_instance') and module in d._instance.dependedBy:
                        self._removeDepend(module, d)
                self._logger.warning('Loading module %r stopped', module, exc_info=True)
                # Not loaded, send a message to notify that parents can not 
                for m in self.waitForSend(ModuleLoadStateChanged(module, ModuleLoadStateChanged.UNLOADED, inst)):
                    yield m
                del module._instance
                raise
            for d in preloads:
                d._instance.dependedBy.add(module)
            self.activeModules[inst.getServiceName()] = module
            for m in module._instance.changestate(ModuleLoadStateChanged.LOADING, self):
                yield m
            for m in module._instance.load(self):
                yield m
            if module._instance.state == ModuleLoadStateChanged.LOADED or module._instance.state == ModuleLoadStateChanged.LOADING:
                sm = ModuleLoadStateChanged.createMatcher(module._instance.target, ModuleLoadStateChanged.SUCCEEDED)
                fm = ModuleLoadStateChanged.createMatcher(module._instance.target, ModuleLoadStateChanged.FAILED)
                yield (sm, fm)
                if self.matcher is sm:
                    pass
                else:
                    raise ModuleLoadException('Module load failed for %r' % (module,))
            self._logger.info('Loading module %r completed, module state is %r', module, module._instance.state)
            if module._instance.state == ModuleLoadStateChanged.FAILED:
                raise ModuleLoadException('Module load failed for %r' % (module,))
    def unloadmodule(self, module, ignoreDependencies = False):
        '''
        Unload a module class. Need delegate to call from other containers.
        '''
        self._logger.debug('Try to unload module %r', module)
        if hasattr(module, '_instance'):
            self._logger.debug('Module %r is loaded, module state is %r', module, module._instance.state)
            inst = module._instance
            if inst.state == ModuleLoadStateChanged.LOADING or inst.state == ModuleLoadStateChanged.LOADED:
                # Wait for loading
                # Wait for succeeded or failed
                sm = ModuleLoadStateChanged.createMatcher(module._instance.target, ModuleLoadStateChanged.SUCCEEDED)
                fm = ModuleLoadStateChanged.createMatcher(module._instance.target, ModuleLoadStateChanged.FAILED)
                yield (sm, fm)
            elif inst.state == ModuleLoadStateChanged.UNLOADING or inst.state == ModuleLoadStateChanged.UNLOADED:
                um = ModuleLoadStateChanged.createMatcher(module, ModuleLoadStateChanged.UNLOADED)
                yield (um,)
        if hasattr(module, '_instance') and (module._instance.state == ModuleLoadStateChanged.SUCCEEDED or
                                             module._instance.state == ModuleLoadStateChanged.FAILED):
            self._logger.info('Unloading module %r', module)
            inst = module._instance
            # Change state to unloading to prevent more dependencies
            for m in inst.changestate(ModuleLoadStateChanged.UNLOADING, self):
                yield m
            if not ignoreDependencies:
                deps = [d for d in inst.dependedBy if hasattr(d, '_instance') and d._instance.state != ModuleLoadStateChanged.UNLOADED]
                ums = [ModuleLoadStateChanged.createMatcher(d, ModuleLoadStateChanged.UNLOADED) for d in deps]
                for d in deps:
                    self.subroutine(self.unloadmodule(d), False)
                while ums:
                    yield ums
                    ums.remove(self.matcher)
            for m in inst.unload(self):
                yield m
            del self.activeModules[inst.getServiceName()]
            self._logger.info('Module %r is unloaded', module)
            if not ignoreDependencies:
                for d in module.depends:
                    if hasattr(d, '_instance') and module in d._instance.dependedBy:
                        self._removeDepend(module, d)
    def loadByPath(self, path):
        """
        Load a module by full path. If there are dependencies, they are also loaded.
        """
        try:
            p, module = findModule(path, True)
        except KeyError as exc:
            raise ModuleLoadException('Cannot load module ' + repr(path) + ': ' + str(exc) + 'is not defined in the package')
        except Exception as exc:
            raise ModuleLoadException('Cannot load module ' + repr(path) + ': ' + str(exc))
        if module is None:
            raise ModuleLoadException('Cannot find module: ' + repr(path))
        for m in self.loadmodule(module):
            yield m
    def unloadByPath(self, path):
        """
        Unload a module by full path. Dependencies are automatically unloaded if they are marked to be
        services.
        """
        p, module = findModule(path, False)
        if module is None:
            raise ModuleLoadException('Cannot find module: ' + repr(path))
        for m in self.unloadmodule(module):
            yield m
    def reloadModules(self, pathlist):
        """
        Reload modules with a full path in the pathlist
        """
        loadedModules = []
        failures = []
        for path in pathlist:
            p, module = findModule(path, False)
            if module is not None and hasattr(module, '_instance') and module._instance.state != ModuleLoadStateChanged.UNLOADED:
                loadedModules.append(module)
        # Unload all modules
        ums = [ModuleLoadStateChanged.createMatcher(m, ModuleLoadStateChanged.UNLOADED) for m in loadedModules]
        for m in loadedModules:
            # Only unload the module itself, not its dependencies, since we will restart the module soon enough
            self.subroutine(self.unloadmodule(m, True), False)
        while ums:
            yield tuple(ums)
            ums.remove(self.matcher)
        # Group modules by package
        grouped = {}
        for path in pathlist:
            dotpos = path.rfind('.')
            if dotpos == -1:
                raise ModuleLoadException('Must specify module with full path, including package name')
            package = path[:dotpos]
            classname = path[dotpos + 1:]
            mlist = grouped.setdefault(package, [])
            p, module = findModule(path, False)
            mlist.append((classname, module))
        for package, mlist in grouped.items():
            # Reload each package only once
            try:
                p = sys.modules[package]
                # Remove cache to ensure a clean import from source file
                removeCache(p)
                p = reload(p)
            except KeyError:
                try:
                    p = __import__(package, fromlist=[m[0] for m in mlist])
                except:
                    self._logger.warning('Failed to import a package: %r, resume others', package, exc_info = True)
                    failures.append('Failed to import: ' + package)
                    continue
            except:
                self._logger.warning('Failed to import a package: %r, resume others', package, exc_info = True)
                failures.append('Failed to import: ' + package)
                continue                
            for cn, module in mlist:
                try:
                    module2 = getattr(p, cn)
                except AttributeError:
                    self._logger.warning('Cannot find module %r in package %r, resume others', package, cn)
                    failures.append('Failed to import: ' + package + '.' + cn)
                    continue
                if module is not None and module is not module2:
                    # Update the references
                    try:
                        lpos = loadedModules.index(module)
                        loaded = True
                    except:
                        loaded = False
                    for d in module.depends:
                        # The new reference is automatically added on import, only remove the old reference
                        d.referencedBy.remove(module)
                        if loaded and hasattr(d, '_instance'):
                            try:
                                d._instance.dependedBy.remove(module)
                                d._instance.dependedBy.add(module2)
                            except ValueError:
                                pass
                    if hasattr(module, 'referencedBy'):
                        for d in module.referencedBy:
                            pos = d.depends.index(module)
                            d.depends[pos] = module2
                            if not hasattr(module2, 'referencedBy'):
                                module2.referencedBy = []
                            module2.referencedBy.append(d)
                    if loaded:
                        loadedModules[lpos] = module2
        # Start the uploaded modules
        for m in loadedModules:
            self.subroutine(self.loadmodule(m))
        if failures:
            raise ModuleLoadException('Following errors occurred during reloading, check log for more details:\n' + '\n'.join(failures))
    def getModuleByName(self, targetname):
        """
        Return the module instance for a target name.
        """
        if targetname == 'public':
            target = None
        elif not targetname not in self.activeModules:
            raise KeyError('Module %r not exists or is not loaded' % (targetname,))
        else:
            target = self.activeModules[targetname]
        return target

def callAPI(container, targetname, name, params = {}, timeout = 60.0):
    """
    Call module API `targetname/name` with parameters. The return value is stored at `container.retvalue`.
    
    :param targetname: module targetname. Usually the lower-cased name of the module class, or 'public' for
                       public APIs.
    
    :param name: method name
    
    :param params: module API parameters, should be a dictionary of `{parameter: value}`
    
    :param timeout: raise an exception if the API call is not returned for a long time
    """
    handle = object()
    apiEvent = ModuleAPICall(handle, targetname, name, params = params)
    for m in container.waitForSend(apiEvent):
        yield m
    replyMatcher = ModuleAPIReply.createMatcher(handle)
    for m in container.waitWithTimeout(timeout, replyMatcher):
        yield m
    if container.timeout:
        # Ignore the Event
        apiEvent.canignore = True
        container.scheduler.ignore(ModuleAPICall.createMatcher(handle))
        raise ModuleAPICallTimeoutException('API call timeout')
    else:
        container.retvalue = getAPIResult(container.event)

def batchCallAPI(container, apis, timeout = 60.0):
    apiHandles = [(object(), api) for api in apis]
    apiEvents = [ModuleAPICall(handle, targetname, name, params = params)
                 for handle, (targetname, name, params) in apiHandles]
    apiMatchers = [ModuleAPIReply.createMatcher(handle) for handle, _ in apiHandles]
    def process():
        for e in apiEvents:
            for m in container.waitForSend(e):
                yield m
    container.subroutine(process(), False)
    eventdict = {}
    def process2():
        while apiMatchers:
            yield tuple(apiMatchers)
            apiMatchers.remove(container.matcher)
            eventdict[container.event.handle] = container.event
    for m in container.executeWithTimeout(timeout, process2()):
        yield m
    for e in apiEvents:
        if e.handle not in eventdict:
            e.canignore = True
            container.scheduler.ignore(ModuleAPICall.createMatcher(e.handle))
    container.retvalue = [eventdict.get(handle, None) for handle, _ in apiHandles]


def getAPIResult(event):
    if hasattr(event, 'exception'):
        raise event.exception
    return event.result



'''
Created on 2016/9/22

:author: hubo
'''

from vlcp.config import Configurable, config
from vlcp.protocol.zookeeper import ZooKeeper, ZooKeeperRetryException,\
    ZooKeeperConnectionStateEvent, ZooKeeperResponseEvent, ZooKeeperWatcherEvent,\
    ZooKeeperSessionExpiredException
from random import shuffle, random, randrange
from vlcp.event.connection import Client
from vlcp.event.event import Event, withIndices
import logging
import vlcp.utils.zookeeper as zk
try:
    from itertools import izip_longest
except Exception:
    izip_longest = zip
from time import time
from vlcp.event.future import RoutineFuture
from contextlib import closing

@withIndices('state', 'client', 'sessionid')
class ZooKeeperSessionStateChanged(Event):
    CREATED = 'created'
    DISCONNECTED = 'disconnected'
    RECONNECTED = 'reconnected'
    AUTHFAILED = 'authfailed'
    EXPIRED = 'expired'


@withIndices('client', 'sessionid', 'restore')
class ZooKeeperRestoreWatches(Event):
    pass


class ZooKeeperSessionUnavailable(Exception):
    def __init__(self, state):
        Exception.__init__(self, "ZooKeeper state is '%r'" % (state,))
        self.state = state


class ZooKeeperIllegalPathException(ValueError):
    pass

_MAX_SETWATCHES_SIZE = 128 * 1024

_should_add_watch = set(((zk.CREATED_EVENT_DEF, zk.ZOO_ERR_NONODE),))

@config('zookeeperclient')
class ZooKeeperClient(Configurable):
    """
    ZooKeeper client to send requests to a cluster
    """
    # Default ZooKeeper server list, should be a list contains connection URLs
    _default_serverlist = []
    # Chroot to a child node instead of the root node. All paths used in the program
    # will be mapped to *chroot_path*/*path*
    _default_chroot = '/'
    # Extra authentications, should be list of tuples [(scheme1, auth1), (scheme2, auth2), ...]
    _default_auth = []
    # Zookeeper session timeout
    _default_sessiontimeout = 20
    # If not None, ZooKeeperClient will disconnect from the server and reconnect to a random server
    # to make sure the connections to ZooKeeper servers are balanced. It sometimes causes problems,
    # so it is disabled by default.
    _default_rebalancetime = None
    _logger = logging.getLogger(__name__ + '.ZooKeeperClient')
    def __init__(self, container, serverlist = None, chroot = None, protocol = None, readonly = False,
                 restart_session = True):
        if serverlist is not None:
            self.serverlist = list(serverlist)
        else:
            self.serverlist = list(self.serverlist)
        shuffle(self.serverlist)
        self.nextptr = 0
        self.current_connection = None
        if protocol is None:
            self.protocol = ZooKeeper()
        else:
            self.protocol = protocol
        self.protocol.persist = False
        if chroot:
            self.chroot = chroot
        else:
            self.chroot = self.chroot
        if not isinstance(self.chroot, bytes):
            self.chroot = self.chroot.encode('utf-8')
        if self.chroot is None or self.chroot == b'/':
            self.chroot = b''
        self.chroot = self.chroot.rstrip(b'/')
        self._container = container
        self.readonly = readonly
        self.auth_set = set(self.auth)
        self.restart_session = restart_session
        self.session_id = 0
        self.session_state = ZooKeeperSessionStateChanged.EXPIRED
        self._shutdown = False
        self.key = None
        self.certificate = None
        self.ca_certs = None
        self._last_zxid = 0
        self._last_watch_zxid = 0
    def start(self, asyncstart = False):
        self._connmanage_routine = self._container.subroutine(self._connection_manage(), asyncstart)
    def reset(self):
        '''
        Discard current session and start a new one
        '''
        self._connmanage_routine.close()
        self._shutdown = False
        self.start()
    def shutdown(self):
        self._connmanage_routine.close()
        if False:
            yield
    def _connection_manage(self):
        try:
            failed = 0
            self._last_zxid = last_zxid = 0
            self._last_watch_zxid = last_watch_zxid = 0
            session_id = 0
            passwd = b'\x00' * 16
            last_conn_time = None
            while True:
                self.currentserver = self.serverlist[self.nextptr]
                np = self.nextptr + 1
                if np >= len(self.serverlist):
                    np = 0
                self.nextptr = np
                conn = Client(self.currentserver, self.protocol, self._container.scheduler,
                              self.key, self.certificate, self.ca_certs)
                self.current_connection = conn
                conn_up = ZooKeeperConnectionStateEvent.createMatcher(ZooKeeperConnectionStateEvent.UP,
                                                                      conn)
                conn_nc = ZooKeeperConnectionStateEvent.createMatcher(ZooKeeperConnectionStateEvent.NOTCONNECTED,
                                                                      conn)
                conn.start()
                try:
                    yield (conn_up, conn_nc)
                    if self._container.matcher is conn_nc:
                        self._logger.warning('Connect to %r failed, try next server', self.currentserver)
                        if failed > 5:
                            # Wait for a small amount of time to prevent a busy loop
                            # Socket may be rejected, it may fail very quick
                            for m in self._container.waitWithTimeout(min((failed - 5) * 0.1, 1.0)):
                                yield m
                        failed += 1
                        continue
                    try:
                        # Handshake
                        set_watches = []
                        if self.session_state == ZooKeeperSessionStateChanged.DISCONNECTED:
                            for m in self._container.waitForSend(ZooKeeperRestoreWatches(self,
                                                                                         self.session_id,
                                                                                         True,
                                                                                         restore_watches = (set(), set(), set()))):
                                yield m
                            yield (ZooKeeperRestoreWatches.createMatcher(self),)
                            data_watches, exists_watches, child_watches = \
                                    self._container.event.restore_watches
                            if data_watches or exists_watches or child_watches:
                                restore_watch_zxid = last_watch_zxid
                                if restore_watch_zxid > 1:
                                    restore_watch_zxid = restore_watch_zxid - 1
                                current_set_watches = zk.SetWatches(relativeZxid = restore_watch_zxid)
                                current_length = 0
                                for d, e, c in izip_longest(data_watches, exists_watches, child_watches):
                                    if d is not None:
                                        current_set_watches.dataWatches.append(d)
                                        current_length += 4 + len(d)
                                    if e is not None:
                                        current_set_watches.existWatches.append(e)
                                        current_length += 4 + len(e)
                                    if c is not None:
                                        current_set_watches.childWatches.append(c)
                                        current_length += 4 + len(c)
                                    if current_length > _MAX_SETWATCHES_SIZE:
                                        # Split set_watches
                                        set_watches.append(current_set_watches)
                                        current_set_watches = zk.SetWatches(relativeZxid = restore_watch_zxid)
                                        current_length = 0
                                if current_set_watches.dataWatches or current_set_watches.existWatches \
                                        or current_set_watches.childWatches:
                                    set_watches.append(current_set_watches)
                        auth_list = list(self.auth_set)
                        with closing(self._container.executeWithTimeout(10,
                                    self.protocol.handshake(conn,
                                        zk.ConnectRequest(lastZxidSeen = last_zxid,
                                                          timeOut = int(self.sessiontimeout * 1000.0),
                                                          sessionId = session_id,
                                                          passwd = passwd,
                                                          readOnly = self.readonly),
                                         self._container,
                                        [zk.AuthPacket(scheme = a[0], auth = a[1]) for a in auth_list] +
                                        set_watches))) as g:
                            for m in g:
                                yield m
                        if self._container.timeout:
                            raise IOError
                    except ZooKeeperSessionExpiredException:
                        self._logger.warning('Session expired.')
                        # Session expired
                        self.session_state = ZooKeeperSessionStateChanged.EXPIRED
                        for m in self._container.waitForSend(ZooKeeperSessionStateChanged(
                                    ZooKeeperSessionStateChanged.EXPIRED,
                                    self,
                                    session_id)):
                            yield m
                        if self.restart_session:
                            failed = 0
                            last_zxid = 0
                            session_id = 0
                            passwd = b'\x00' * 16
                            last_conn_time = None
                            continue
                        else:
                            break                        
                    except Exception:
                        self._logger.warning('Handshake failed to %r, try next server', self.currentserver)
                        if failed > 5:
                            # There is a bug ZOOKEEPER-1159 that ZooKeeper server does not respond
                            # for session expiration, but directly close the connection.
                            # This is a workaround: we store the time that we disconnected from the server,
                            # if we have exceeded the session expiration time, we declare the session is expired
                            if last_conn_time is not None and last_conn_time + self.sessiontimeout * 2 < time():
                                self._logger.warning('Session expired detected from client time.')
                                # Session expired
                                self.session_state = ZooKeeperSessionStateChanged.EXPIRED
                                for m in self._container.waitForSend(ZooKeeperSessionStateChanged(
                                            ZooKeeperSessionStateChanged.EXPIRED,
                                            self,
                                            session_id)):
                                    yield m
                                if self.restart_session:
                                    failed = 0
                                    last_zxid = 0
                                    session_id = 0
                                    passwd = b'\x00' * 16
                                    last_conn_time = None
                                    continue
                                else:
                                    break                        
                            else:
                                # Wait for a small amount of time to prevent a busy loop
                                for m in self._container.waitWithTimeout(min((failed - 5) * 0.1, 1.0)):
                                    yield m
                        failed += 1
                    else:
                        failed = 0
                        conn_resp, auth_resp = self._container.retvalue
                        if conn_resp.timeOut <= 0:
                            # Session expired
                            # Currently should not happen because handshake() should raise an exception
                            self._logger.warning('Session expired detected from handshake packet')
                            self.session_state = ZooKeeperSessionStateChanged.EXPIRED
                            for m in self._container.waitForSend(ZooKeeperSessionStateChanged(
                                        ZooKeeperSessionStateChanged.EXPIRED,
                                        self,
                                        session_id)):
                                yield m
                            if self.restart_session:
                                failed = 0
                                last_zxid = 0
                                last_conn_time = None
                                session_id = 0
                                passwd = b'\x00' * 16
                                continue
                            else:
                                break
                        else:
                            session_id = conn_resp.sessionId
                            passwd = conn_resp.passwd
                            # Authentication result check
                            auth_failed = any(a.err == zk.ZOO_ERR_AUTHFAILED for a in auth_resp)
                            if auth_failed:
                                self._logger.warning('ZooKeeper authentication failed for following auth: %r',
                                                     [a for a,r in zip(auth_list, auth_resp) if r.err == zk.ZOO_ERR_AUTHFAILED])
                                self.session_state = ZooKeeperSessionStateChanged.AUTHFAILED
                                for m in self._container.waitForSend(ZooKeeperSessionStateChanged(
                                                ZooKeeperSessionStateChanged.AUTHFAILED,
                                                self,
                                                session_id
                                            )):
                                    yield m
                                # Not retrying
                                break
                            else:
                                self.session_readonly = getattr(conn_resp, 'readOnly', False)
                                self.session_id = session_id
                                if self.session_state == ZooKeeperSessionStateChanged.EXPIRED:
                                    for m in self._container.waitForSend(ZooKeeperSessionStateChanged(
                                                ZooKeeperSessionStateChanged.CREATED,
                                                self,
                                                session_id
                                            )):
                                        yield m
                                else:
                                    for m in self._container.waitForSend(ZooKeeperSessionStateChanged(
                                                ZooKeeperSessionStateChanged.RECONNECTED,
                                                self,
                                                session_id
                                            )):
                                        yield m
                                self.session_state = ZooKeeperSessionStateChanged.CREATED
                        if conn.connected:
                            conn_down = ZooKeeperConnectionStateEvent.createMatcher(ZooKeeperConnectionStateEvent.DOWN,
                                                                                    conn,
                                                                                    conn.connmark
                                                                                    )
                            auth_failed = ZooKeeperResponseEvent.createMatcher(zk.AUTH_XID,
                                                                               conn,
                                                                               conn.connmark,
                                                                               _ismatch = lambda x: x.message.err == ZOO_ERR_AUTHFAILED)
                            while True:
                                rebalancetime = self.rebalancetime
                                if rebalancetime is not None:
                                    rebalancetime += random() * 60
                                for m in self._container.waitWithTimeout(rebalancetime, conn_down, auth_failed):
                                    yield m
                                if self._container.timeout:
                                    # Rebalance
                                    if conn.zookeeper_requests:
                                        # There are still requests not processed, wait longer
                                        for _ in range(0, 3):
                                            longer_time = random() * 10
                                            for m in self._container.waitWithTimeout(longer_time, conn_down, auth_failed):
                                                yield m
                                            if not self._container.timeout:
                                                # Connection is down, or auth failed
                                                break
                                            if not conn.zookeeper_requests:
                                                break
                                        else:
                                            # There is still requests, skip for this time
                                            continue
                                    # Rebalance to a random server
                                    if self._container.timeout:
                                        self.nextptr = randrange(len(self.serverlist))
                                break
                            if self._container.matcher is auth_failed:
                                self._logger.warning('ZooKeeper authentication failed, shutdown the connection')
                                self.session_state = ZooKeeperSessionStateChanged.AUTHFAILED
                                for m in self._container.waitForSend(ZooKeeperSessionStateChanged(
                                                ZooKeeperSessionStateChanged.AUTHFAILED,
                                                self,
                                                session_id
                                            )):
                                    yield m
                                # Not retrying
                                break
                            else:
                                # Connection is down, try other servers
                                if not self._container.timeout:
                                    self._logger.warning('Connection lost to %r, try next server', self.currentserver)
                                else:
                                    self._logger.info('Rebalance to next server')
                                self._last_zxid = last_zxid = conn.zookeeper_lastzxid
                                self._last_watch_zxid = last_watch_zxid = conn.zookeeper_last_watch_zxid
                                last_conn_time = time()
                                self.session_state = ZooKeeperSessionStateChanged.DISCONNECTED
                                for m in self._container.waitForSend(ZooKeeperSessionStateChanged(
                                                ZooKeeperSessionStateChanged.DISCONNECTED,
                                                self,
                                                session_id
                                            )):
                                    yield m                                    
                finally:
                    conn.subroutine(conn.shutdown(True), False)
                    self.current_connection = None
        finally:
            self._shutdown = True
            if self.session_state != ZooKeeperSessionStateChanged.EXPIRED and self.session_state != ZooKeeperSessionStateChanged.AUTHFAILED:
                self.session_state = ZooKeeperSessionStateChanged.EXPIRED
                self._container.scheduler.emergesend(ZooKeeperSessionStateChanged(
                                                ZooKeeperSessionStateChanged.EXPIRED,
                                                self,
                                                session_id
                                            ))
    def chroot_path(self, path):
        return self.chroot + path
    def unchroot_path(self, path):
        return path[len(self.chroot):]
    def _analyze(self, request):
        if request.type in (zk.ZOO_EXISTS_OP, zk.ZOO_GETDATA_OP, zk.ZOO_GETACL_OP, zk.ZOO_GETCHILDREN_OP,
                            zk.ZOO_SYNC_OP, zk.ZOO_PING_OP, zk.ZOO_GETCHILDREN2_OP, zk.ZOO_SETAUTH_OP):
            # These requests can be retried even if they are already sent 
            can_retry = True
        else:
            can_retry = False
        watch_type = None
        if request.type == zk.ZOO_MULTI_OP:
            # chroot sub ops
            for op in request.requests:
                if hasattr(op, 'path'):
                    op.path = self.chroot_path(op.path)
        else:
            if hasattr(request, 'path'):
                request.path = self.chroot_path(request.path)
            if getattr(request, 'watch', False):
                if request.type == zk.ZOO_GETDATA_OP:
                    watch_type = zk.CHANGED_EVENT_DEF
                elif request.type == zk.ZOO_EXISTS_OP:
                    watch_type = zk.CREATED_EVENT_DEF
                elif request.type == zk.ZOO_GETCHILDREN_OP or request.type == zk.ZOO_GETCHILDREN2_OP:
                    watch_type = zk.CHILD_EVENT_DEF
        return (request, can_retry, watch_type)
    def watch_path(self, path, watch_type, container):
        '''
        Watch the specified path as specified type
        '''
        if watch_type == zk.CHANGED_EVENT_DEF:
            watch_matchers = (ZooKeeperWatcherEvent.createMatcher(None, None, self.protocol, zk.CHANGED_EVENT_DEF, None, path),
                              ZooKeeperWatcherEvent.createMatcher(None, None, self.protocol, zk.DELETED_EVENT_DEF, None, path))
        else:
            watch_matchers = (ZooKeeperWatcherEvent.createMatcher(None, None, self.protocol, watch_type, None, path),)
        # If the session expires, raise exception and exit
        session_state = ZooKeeperSessionStateChanged.createMatcher(ZooKeeperSessionStateChanged.EXPIRED,
                                                                   self,
                                                                   self.session_id)
        auth_failed = ZooKeeperSessionStateChanged.createMatcher(ZooKeeperSessionStateChanged.AUTHFAILED,
                                                                   self,
                                                                   self.session_id)
        # If the watchers are restored, restore the matchers
        restore_matcher = ZooKeeperRestoreWatches.createMatcher(self, self.session_id, True)
        while True:
            yield (session_state, auth_failed, restore_matcher) + watch_matchers
            if container.matcher is session_state or container.matcher is auth_failed:
                raise ZooKeeperSessionUnavailable(container.event.state)
            elif container.matcher is restore_matcher:
                ev = container.event
                ev.restore_watches[{zk.CHANGED_EVENT_DEF : 0,
                                   zk.CREATED_EVENT_DEF : 1,
                                   zk.CHILD_EVENT_DEF : 2}[watch_type]].add(path)
            else:
                watcher_event = container.event.message
                if watcher_event.path:
                    watcher_event.path = self.unchroot_path(watcher_event.path)
                container.retvalue = watcher_event
                break
    def requests(self, requests, container, timeout = None, session_lock = None, callback = None, priority = 0):
        '''
        similar to vlcp.protocol.zookeeper.ZooKeeper.requests, but:
           1. Returns an extra item *watchers*, which is a list of objects corresponding to each request.
              if the request has watch=True, the corresponding object is a RoutineFuture object;
              if the request has watch=False or does not support watch, the corresponding object is None.
              Use watcher.wait() to get the watch event. Use watcher.close() to discard the watcher.
          
           2. If the connection is lost during requests, this method waits for reconnecting until timeout, 
              session expires or the response of a request which is not read-only is lost.
        
        :param requests: sequence of request
        
        :param container: container of current routine
        
        :param timeout: if not None, wait only for specified time. Notice that it is not an exact limit,
                it won't stop the execution unless the connection is lost
        
        :param session_lock: if not None, only execute if the session_id == session_lock
        
        :param callback: if not None, callback(request, response) is called immediately after any response is received
        
        :return: (result, lost_responses, retry_requests, watchers) tuple, the first three are the
                same as ZooKeeper.requests, the last item *watchers* is a list of RoutineFuture objects
        '''
        if self._shutdown:
            raise ZooKeeperSessionUnavailable(self.session_state)
        if session_lock is not None and self.session_id != session_lock:
            raise ZooKeeperSessionUnavailable(ZooKeeperSessionStateChanged.EXPIRED)
        start_time = time()
        if timeout is not None:
            end_time = start_time + timeout
        def left_time():
            if timeout is None:
                return None
            else:
                return max(end_time - time(), 0)
        def has_time_left():
            t = left_time()
            return t is None or t > 0
        result = {}
        lost_responses = []
        analysis = dict((v[0], (v[1], v[2])) for v in (self._analyze(r) for r in requests))
        retry_requests = list(requests)
        watchers = {}
        def requests_callback(request, response):
            watch_type = analysis[request][1]
            if watch_type is not None and (response.err == zk.ZOO_ERR_OK or \
                                           (watch_type, response.err) in _should_add_watch):
                watchers[request] = RoutineFuture(self.watch_path(request.path, watch_type, container), container)
            if callback is not None:
                callback(request, response)
        def unchroot_response(resp):
            if resp.zookeeper_request_type == zk.ZOO_MULTI_OP:
                for r in resp.responses:
                    if hasattr(r, 'path'):
                        r.path = self.unchroot_path(r.path)
            elif hasattr(resp, 'path'):
                resp.path = self.unchroot_path(resp.path)
            return resp
        while has_time_left() and not lost_responses and retry_requests:
            if self.session_state != ZooKeeperSessionStateChanged.CREATED:
                def wait_for_connect():
                    state_change = ZooKeeperSessionStateChanged.createMatcher(None, self)
                    while True:
                        yield (state_change,)
                        if container.event.state in (ZooKeeperSessionStateChanged.CREATED, ZooKeeperSessionStateChanged.RECONNECTED):
                            break
                        elif self._shutdown:
                            raise ZooKeeperSessionUnavailable(self.session_state)
                        elif session_lock is not None and (container.event.sessionid != session_lock or \
                                                           container.event.state == ZooKeeperSessionStateChanged.EXPIRED):
                            raise ZooKeeperSessionUnavailable(ZooKeeperSessionStateChanged.EXPIRED)
                try:
                    for m in container.executeWithTimeout(left_time(), wait_for_connect()):
                        yield m
                except ZooKeeperSessionUnavailable:
                    if len(retry_requests) == len(requests):
                        raise
                    else:
                        break
                if container.timeout:
                    if len(retry_requests) == len(requests):
                        raise ZooKeeperSessionUnavailable(ZooKeeperSessionStateChanged.DISCONNECTED)
                    else:
                        break
            # retry all the requests
            for m in self.protocol.requests(self.current_connection, retry_requests, container,
                                            requests_callback, priority=priority):
                yield m
            new_result, new_lost, new_retry = container.retvalue
            # Save the results
            result.update((k,unchroot_response(v)) for k,v in zip(retry_requests, new_result) if v is not None)
            if new_lost:
                # Some responses are lost
                for i in range(len(new_lost) - 1, -1, -1):
                    if not analysis[new_lost[i]][0]:
                        # This request can not be retried
                        break
                else:
                    i = -1
                new_retry = new_lost[i+1:] + new_retry
                new_lost = new_lost[:i+1]
                if new_lost:
                    # Some requests can not be retried, this is as far as we go
                    lost_responses = new_lost
                    retry_requests = new_retry
                    break
            retry_requests = new_retry
        container.retvalue = ([result.get(r, None) for r in requests],
                              lost_responses,
                              retry_requests,
                              [watchers.get(r, None) for r in requests])
    def get_last_zxid(self):
        '''
        Return the latest zxid seen from servers
        '''
        if not self.current_connection:
            return self._last_zxid
        else:
            return getattr(self.current_connection, 'zookeeper_lastzxid', self._last_zxid)
    def get_last_watch_zxid(self):
        '''
        Return the latest zxid seen from servers
        '''
        if not self.current_connection:
            return self._last_watch_zxid
        else:
            return getattr(self.current_connection, 'zookeeper_last_watch_zxid', self._last_watch_zxid)
